"""
jbonet @ boliva's lab 2013
"""
import  numpy               as np
from    numpy import linalg as LA
import  copy, sys

class SecondaryStructure(object):

    min_ss_length     = {'H':4, 'E':2, 'G':3}
    max_ini_distance  = {'H':8, 'E':2, 'G':6}
    structure_regions = set(['H','G','E'])

    def __init__(self, sstype, initposition):

        self._sstype                                        = sstype
        self._inip                                          = initposition
        self._struct                                        = None
        self._endp                                          = None
        self._length                                        = None
        self._f11, self._cmf11, self._eigf11, self._tsf11   = None, None, None, None
        self._f44, self._cmf44, self._eigf44, self._tsf44   = None, None, None, None

    @property
    def structure(self):
        return self._struct
        
    def get_moment_of_inertia_length(self, workp):
        if   workp == 'f11':    (p1, p2) = (self._get_coordinate(self._f11),  self._get_coordinate(self._endp))
        elif workp == 'f44':    (p1, p2) = (self._get_coordinate(self._inip), self._get_coordinate(self._f44) )
        return p2 - p1 + 1

    def _get_coordinate(self, identifier):
        return self._struct._get_structure_array_coordinate(identifier)

    def calculate_center_of_masses(self):
        _struct = self._struct.duplicate(backbone = True)
        if self.max_ini_distance[self._sstype] >= self._length:
            self._f11   = self._inip
            self._f44   = self._endp
            self._cmf11 = _struct.geometric_center()
            self._cmf44 = self._cmf11
        else:
            self._f11   = _struct.aminoacids[ len(_struct.aminoacids) - self.max_ini_distance[self._sstype] ].identifier
            self._f44   = _struct.aminoacids[ self.max_ini_distance[self._sstype] - 1 ].identifier
            self._cmf11 = _struct.extract(self._f11, self._endp).geometric_center()
            self._cmf44 = _struct.extract(self._inip, self._f44).geometric_center()

        self._jacobi_angles('f11')
        if self._f11 != self._inip and self._f44 != self._endp:
            self._jacobi_angles('f44')
            if self._sstype == 'E': self._process_betas()
        else:
            self._eigf44 = self._eigf11 

        self._orientvectors()

    def _orientvectors(self):
        first_halfdif   = np.subtract(self._struct.first_aminoacid.ca.coordinates, self._cmf44)
        second_halfdif  = np.subtract(self._struct.last_aminoacid.ca.coordinates,  self._cmf11)
    
        first_sign      = -1 * np.sign(np.dot(first_halfdif,  self._eigf44))
        second_sign     = np.sign(np.dot(second_halfdif, self._eigf11))

        self._eigf44    = np.multiply(self._eigf44, first_sign)
        self._eigf11    = np.multiply(self._eigf11, second_sign)

        first_lambda    = np.dot(first_halfdif,  self._eigf44)
        second_lambda   = np.dot(second_halfdif, self._eigf11)

        self._tsf44     = np.add(np.multiply(first_lambda,  self._eigf44), self._cmf44)
        self._tsf11     = np.add(np.multiply(second_lambda, self._eigf11), self._cmf11)

    def _jacobi_angles(self, workp):
        _struct = self._struct.duplicate(backbone = True)
        if workp == 'f11':
            moving_point = self._f11
            fixed_point  = self._endp
            ini_coord    = _struct.extract(self._f11, self._endp)._all_atoms_coordinates()
            distance     = len(ini_coord)/3
            cm           = self._cmf11
        elif workp == 'f44':
            moving_point = self._f44
            ini_coord    = _struct.extract(self._inip, self._f44)._all_atoms_coordinates()
            distance     = len(ini_coord)/3
            cm           = self._cmf44

        new_coord = np.subtract(ini_coord, cm)
        x2, y2, z2, xy, yz, zx = 0, 0, 0, 0, 0, 0
        if self._sstype != 'E':
            for row in new_coord:
                x, y, z = row
                x2 += np.power(x, 2)
                y2 += np.power(y, 2)
                z2 += np.power(z, 2)
                xy += np.multiply(x, y)
                yz += np.multiply(y, z)
                zx += np.multiply(z, x)
        else:
            for i in range(0, len(ini_coord) - 4, 3):
                j=i+3
                x0, y0, z0 = new_coord[i:i+3][:,0], new_coord[i:i+3][:,1], new_coord[i:i+3][:,2]
                x1, y1, z1 = new_coord[j:j+3][:,0], new_coord[j:j+3][:,1], new_coord[j:j+3][:,2]

                xm, ym, zm = np.mean([x0, x1], axis = 0), np.mean([y0, y1], axis = 0), np.mean([z0, z1], axis = 0)

                x2 += np.sum(np.power(xm, 2))
                y2 += np.sum(np.power(ym, 2))
                z2 += np.sum(np.power(zm, 2))
                xy += np.sum(np.multiply(xm, ym))
                yz += np.sum(np.multiply(ym, zm))
                zx += np.sum(np.multiply(zm, xm))

        a = np.matrix([[y2 + z2,     -xy,     -zx],
                       [    -xy, x2 + z2,     -yz],
                       [    -zx,     -yz, x2 + y2]])

        eigenVal, eigenVec = LA.eig(a)
        #sort eigenVal / eigenVec by descending eigenVal value
        #get the smallest value as represents the axis that actually follows the structure direction
        idx = eigenVal.argsort()[::-1]
        eigenVal = eigenVal[idx]
        eigenVec = eigenVec[:,idx]

        if   workp == 'f11':    self._eigf11 = np.asarray(eigenVec[:,2]).reshape(-1)
        elif workp == 'f44':    self._eigf44 = np.asarray(eigenVec[:,2]).reshape(-1)

    def _process_betas(self):
        advance_limit        = 2 #plus the two already selected makes a total of 4
        degree_dif_threshold = 10

        end, advance        = False, 1
        original_eigienf11  = self._eigf11
        while not end:
            new_me          = copy.deepcopy(self)
            _struct         = new_me._struct.duplicate(backbone = True)
            new_me._f11     = new_me._struct.aminoacids[ len(new_me._struct.aminoacids) - (self.max_ini_distance[new_me._sstype]  + advance)].identifier
            new_me._cmf11   = _struct.extract(new_me._f11, new_me._endp).geometric_center()

            new_me._jacobi_angles('f11')
            difference = np.degrees(np.arccos(np.absolute(np.dot(new_me._eigf11, original_eigienf11))))
            if new_me._f11 == new_me._inip or advance >= advance_limit: end = True
            advance += 1
            if difference < degree_dif_threshold:
                self._f11    = new_me._f11
                self._cmf11  = new_me._cmf11
                self._eigf11 = new_me._eigf11

        end, advance        = False, 1
        original_eigienf44  = self._eigf44
        while not end:
            new_me          = copy.deepcopy(self)
            _struct         = new_me._struct.duplicate(backbone = True)
            new_me._f44     = new_me._struct.aminoacids[ self.max_ini_distance[new_me._sstype] - 1 + advance].identifier
            new_me._cmf44   = _struct.extract(new_me._inip, new_me._f44).geometric_center()

            new_me._jacobi_angles('f44')
            difference = np.degrees(np.arccos(np.absolute(np.dot(original_eigienf44, new_me._eigf44))))
            if new_me._f44 == new_me._endp or advance >= advance_limit: end = True
            advance += 1
            if difference < degree_dif_threshold:
                self._f44    = new_me._f44
                self._cmf44  = new_me._cmf44
                self._eigf44 = new_me._eigf44

    def headerformat(self, ssnum = 1, ssidentifier = 'A'):
        inires = self._struct.aminoacids[0]
        endres = self._struct.aminoacids[-1]
        if self._sstype == 'H' or self._sstype == 'G':
            ini = '{1.type} {0._struct.chain}{1.number:>5d}{1.version}'.format(self, inires)
            end = '{1.type} {0._struct.chain}{1.number:>5d}{1.version}'.format(self, endres)
            return 'HELIX {0:>4d} {1:>3s} {2} {3} 1 {4:>35d}'.format(ssnum, ssidentifier, ini, end, self._length)
        elif self._sstype == 'E':
            ini = '{1.type} {0._struct.chain}{1.number:>4d}{1.version}'.format(self, inires)
            end = '{1.type} {0._struct.chain}{1.number:>4d}{1.version}'.format(self, endres)
            return 'SHEET {0:>4d} {1:>3s} 1 {2} {3} 0'.format(ssnum, ssidentifier, ini, end)
        else:
            raise NotImplementedError('Unknown secondary structure type!')

    def strdata(self, workp):
        data = []
        data.append("( {0._sstype} ) {0._inip:>4s} <-- {0._length:>2d} --> {0._endp:>4s}".format(self))
        if workp == 'f11':
            data.append("\tf11: {0._f11!s:>4s} cmf11: {0._cmf11!s:>50s} eigf11: {0._eigf11!s:>50s}".format(self))
        elif workp == 'f44':
            data.append("\tf44: {0._f44!s:>4s} cmf44: {0._cmf44!s:>50s} eigf44: {0._eigf44!s:>50s}".format(self))
        return "\n".join(data)

    def __repr__(self):
        data = []
        data.append("( {0._sstype} ) {0._inip:>4s} <-- {0._length:>2d} --> {0._endp:>4s}".format(self))
        data.append("\tf11: {0._f11:>4s} cmf11: {0._cmf11:>50s} eigf11: {0._eigf11:>50s}".format(self))
        data.append("\tf44: {0._f44:>4s} cmf44: {0._cmf44:>50s} eigf44: {0._eigf44:>50s}".format(self))
        # data.append(repr(self._struct))
        return "\n".join(data)



"""
PHInnerContact

author: jbonet
date:   03/2013

@oliva's lab
"""
from ..interface import Interface
from ..contact   import ContactAH

class PHInnerContact(Interface):
    """
    {Interface} between two protein chains
    """
    def __init__(self, protein = None, threshold_type = "min", threshold_distance = 8):

        protein_centers = protein.geometric_center(structure = True,  hetero     = False,
                                                   water     = False, by_residue = True)
        hetero_centers  = protein.geometric_center(structure = False, hetero     = True,
                                                   water     = False, by_residue = True)

        super(PHInnerContact, self).__init__(chain1         = protein,         chain2             = protein,
                                             threshold_type = threshold_type,  threshold_distance = threshold_distance,
                                             chain1_centres = protein_centers, chain2_centers     = hetero_centers)
        
    #
    # ATTRIBUTES
    #
    @property
    def protein(self):       return self._chain1
    @property
    def protein_id(self):    return self._chain1.globalID

    @property
    def protein_positions(self):
        return self._list_positions(1)
    @property 
    def protein_view_innercontact(self):
        return self._view_interface_from(1)

    @property
    def heteroatoms_positions(self):
        return self._list_positions(2)
    @property 
    def heteroatoms_view_innercontact(self):
        return self._view_interface_from(2)

    #
    # OVERWRITE PARENT: METHODS
    #
    @staticmethod
    def test_identifier(chain):
        return Interface.test_identifier(chain, chain)

    #
    # OVERWRITE PARENT: PRIVATE FUNCTIONS
    #
    def _build(self):
        if len(self.protein.aminoacids) == 0 or len(self.protein.heteroatoms) == 0: return

        super(PHInnerContact, self)._build()

        for i in range(len(self._filtered[0])):
            new_contact = ContactAH(aminoacid          = self.protein.aminoacids[self._filtered[0][i]], 
                                    heteroatom         = self.protein.heteroatoms[self._filtered[1][i]], 
                                    threshold_type     = self.threshold_type, 
                                    threshold_distance = self.threshold_distance)
            if new_contact.is_underthreshold:
                self.contacts = new_contact

    #
    # OVERWRITE PARENT: TOSTRING
    #
    def toString(self, all_types = False):
        data = []
        if not all_types:
            data.append('{0._chain1.chain}\t{0._chain2.chain}\t{0.threshold_type}\t{0.threshold_distance}'.format(self))
        else:
            data.append('{0._chain1.chain}\t{0._chain2.chain}\tmin\tgeometric'.format(self))
        data.append(super(PHInnerContact, self).toString(all_types))
        return "\n".join(data)


from collections import OrderedDict
from SBILib.data       import aminoacids3to1, nucleic3to1
from .chain         import Chain, ChainOfProtein, ChainOfNucleotide
import gzip
import re

columns_order = [
    "group_PDB",
    "id",
    "type_symbol",
    "label_atom_id",
    "label_alt_id",
    "label_comp_id",
    "label_asym_id",
    "label_entity_id",
    "label_seq_id",
    "pdbx_PDB_ins_code",
    "Cartn_x",
    "Cartn_y",
    "Cartn_z",
    "occupancy",
    "B_iso_or_equiv",
    "pdbx_formal_charge",
    "auth_seq_id",
    "auth_comp_id",
    "auth_asym_id",
    "auth_atom_id",
    "pdbx_PDB_model_num"
]


def specialSplit( content, monoliners ):
    output = [["", False]]
    quote  = False
    length = len(content)

    if re.search(r'^(?:"\w*)\'\S{1}\'^(?:"\w*)', content) and not content.startswith('_'):
        output[-1][0] == content
        return output

    if monoliners:
        if content.startswith(';') and content.endswith(';'):
            output[-1][0] = content.strip(';')
            return output

    for c in range(length):
        isWS   = content[c] == " " or content[c] == "\t"
        wasWS  = c == 0 or content[c - 1] == " " or content[c - 1] == "\t"
        willWS = c == length - 1 or content[c + 1] == " " or content[c + 1] == "\t"
        braket = frozenset(["'", '"'])
        if (content[c] in braket) and (wasWS or willWS):
            quote = not quote
        elif not quote and isWS and output[-1][0] != "":
            output.append(["", False])
        elif not quote and content[c] == "#":
            # Beware of identifiers with # sign!
            # Problem PBD ID: 2XQB
            if c == 0 or content[c - 1] == ' ':
                break
        elif not isWS or quote:
            output[-1][0] += content[c]
            output[-1][1] = quote
    if output[-1][0] == "":
        output.pop()
    return output


def typefy( content, only_empties=False ):
    if content == "?":
        return " "
    if content == ".":
        return " "

    if only_empties:
        return content

    if "." in content:
        try:
            return float(content)
        except Exception:
            pass
    else:
        try:
            return int(content)
        except Exception:
            pass
    return content


class mmCIF( object ):
    """Read mmCIF format as defined in http://mmcif.wwpdb.org/ and transform it
    into a json object.
    This parser is based on the one created by Gert-Jan Bekker
    <http://github.com/gjbekker/cif-parsers>, but highly simplified.
    """
    def __init__( self, pdbobject, monoliners=False ):
        self.multiline  = False
        self.buffer     = ""
        self.data       = OrderedDict()
        self.loopstatus = False
        self.keyname    = None
        self.lastkey    = None
        self.lastsubkey = None
        self.headers    = frozenset(['loop_', 'save_', 'global_', 'data_'])
        self.monoliners = monoliners
        self.WriteCIF = True

    def read( self, pdbobject ):
        """Parse the provided file.
        :param str file_name: Name of the file to parse.
        :return: json containing the file's data
        """
        file_name = pdbobject._cif_file.name
        fd = gzip.open(file_name, 'rt') if file_name.endswith("gz") else open(file_name)
        return self._from_filehandle(fd, pdbobject)

    def _from_filehandle( self, fhandle, pdbobject ):
        for line in [l.rstrip() for l in fhandle]:
            if len(line) == 0:
                continue
            if self._check_multiline(line):
                self._assign(specialSplit(self.buffer, self.monoliners), pdbobject)
                self.buffer = ""
        info = self.data
        self._set_default(True)
        return info

    def _is_header( self, data ):
        for x in self.headers:
            if data[0].startswith(x) and not data[1]:
                return True
        return False

    def _manage_header( self, data ):
        if data[0][0] == 'loop_':
            self.loopstatus = True
        elif data[0][0].startswith('data_'):
            self.keyname = data[0][0][5:].strip()
            self.data.setdefault(self.keyname, {})
        elif data[0][0].startswith('save_'):
            pass  # TODO
        elif data[0][0].startswith('global_'):
            pass  # TODO

    def _set_default( self, strong=False ):
        self.loopstatus = False
        self.lastkey    = None
        if strong:
            self.multiline  = False
            self.buffer     = ""
            self.data       = OrderedDict()
            self.keyname    = None

    def _assign( self, data, pdbobject ):
        if len(data) == 0:
            self._set_default()
        elif self._is_header(data[0]):
            self._manage_header(data)
        elif not self.loopstatus:
            self._no_loop_status(data, pdbobject)
        else:
            self._loop_status(data, pdbobject)

    def _loop_status( self, data, pdbobject ):
        if data[0][0].startswith('_'):
            k = data[0][0].split('.')
            if self.lastkey is None:
                self.lastkey = {}
            self.lastkey.setdefault(k[0], []).append(k[1])
            self.data[self.keyname].setdefault(k[0], OrderedDict()).setdefault(k[1], [])
            self.lastsubkey = 0
        else:
            k = list(self.lastkey.keys())[0]
            for x in range(len(data)):
                #try:
                """
                subtracted 1 from [self.lastsubkey + x] for proper indexing
                """
                self.data[self.keyname][k][self.lastkey[k][self.lastsubkey + x - 1]].append(typefy(data[x][0], True))
                #except:
                     #print(len(data))
                     #print(x)
                     #exit(0)
                """
                -Patrick Gohl
                """
            self.lastsubkey = self.lastsubkey + x + 1
            
            if k == '_pdbx_poly_seq_scheme' and self.WriteCIF:
                self._write_to_pdb_object(pdbobject)
                self.WriteCIF = False
            if self.lastsubkey >= len(self.lastkey[k]):
                self.lastsubkey = 0

    def _no_loop_status( self, data, pdbobject ):
        if data[0][0].startswith('_'):
            self.lastkey = data[0][0].split('.')
            self.data[self.keyname].setdefault(self.lastkey[0], OrderedDict()).setdefault(self.lastkey[1], [])
            if len(data) == 2:
                d = typefy(data[1][0], True)
                self.data[self.keyname][self.lastkey[0]][self.lastkey[1]].append(d)
                self.lastkey = None
        elif self.lastkey is not None:
            if len(data) == 1:
                d = typefy(dataprotein_key[0][0], True)
                self.data[self.keyname][self.lastkey[0]][self.lastkey[1]].append(d)
            else:
                d = typefy(" ".join([x[0] for x in data]))
                self.data[self.keyname][self.lastkey[0]][self.lastkey[1]].append(d)
            self.lastkey = None

    # def _check_multiline( self, line ):
    #     if line[0] == ';':
    #         self.multiline = not self.multiline
    #     if self.multiline or line.strip() == ';':
    #         self.buffer += line  # .lstrip(';')
    #         return False
    #     else:
    #         if len(self.buffer) == 0:
    #             self.buffer = line
    #         return True

    def _check_multiline( self, line ):
        if line[0] == ';':
            self.buffer += ';'
            self.multiline = not self.multiline
        if self.multiline:
            if not self.monoliners:
                self.buffer += line.lstrip(';')
            else:
                self.buffer += line
            return False
        else:
            if len(self.buffer) == 0:
                self.buffer = line
            return True

    def _write_to_pdb_object(self, pdbobject):
            for key in list(self.data.keys()):
                old_chain = "Old_Chain"
                
                for index in range(len(self.data[key]['_atom_site']["label_entity_id"])):
                    """
                    The ordering is off in the column keys, everything was shifted one position to the right 
                    """
                    pos1 =  self.data[key]['_atom_site']['pdbx_PDB_model_num'][index]  
                    pos2 = self.data[key]['_atom_site']['group_PDB'][index]
                    pos3 = self.data[key]['_atom_site']['type_symbol'][index]
                    pos4 = self.data[key]['_atom_site']['label_alt_id'][index]
                    pos5 = self.data[key]['_atom_site']['auth_comp_id'][index]
                    pos6 = self.data[key]['_atom_site']['pdbx_formal_charge'][index]
                    pos7 = self.data[key]['_atom_site']['pdbx_PDB_ins_code'][index]
                    pos8 = self.data[key]['_atom_site']['Cartn_x'][index]
                    pos9 = self.data[key]['_atom_site']['Cartn_y'][index]
                    pos10 = self.data[key]['_atom_site']['Cartn_z'][index]
                    pos11 = self.data[key]['_atom_site']['occupancy'][index]
                    pos12 = self.data[key]['_atom_site']['id'][index]
                    pos13 = self.data[key]['_atom_site']['B_iso_or_equiv'][index]
                    pdbline = pos1.ljust(6) + pos2.rjust(5) + "  " + pos3.ljust(4) + pos4.rjust(3) + pos5.rjust(2) + pos6.rjust(4) + "    " + pos7.rjust(8) + pos8.rjust(8) + pos9.rjust(8) + pos10.rjust(6) + pos11.rjust(6) + pos12.rjust(12) + pos13.rjust(2)
  
                    chain = self.data[key]['_atom_site']['auth_comp_id'][index]
                    
                    if self.data[key]['_atom_site']['auth_atom_id'][index] != "1":
                        old_chain = "Old_Chain"
                        pdbobject._NMR = True
                    if chain != old_chain:
                        chain_type = self.data[key]['_atom_site']['label_alt_id'][index]
                        if chain_type in aminoacids3to1:
                            obj_chain = ChainOfProtein(pdb = key ,chain = chain)
                            pdbobject._has_prot = True
                        elif chain_type in nucleic3to1:
                            obj_chain = ChainOfNucleotide(pdb = key ,chain = chain)
                            pdbobject._has_nucl = True
                        else:
                            obj_chain = Chain(pdb = key ,chain = chain)
                            pdbobject.add_chain(obj_chain, NMR = pdbobject.is_NMR) 
                    else:
                        putative_old_chain = pdbobject.get_chain_by_id(id = chain)
                        obj_chain = putative_old_chain
                    
                    old_chain = chain

                    chain_type = self.data[key]['_atom_site']['label_alt_id'][index]
                    if not isinstance(obj_chain, ChainOfNucleotide) and not isinstance(obj_chain, ChainOfProtein):
                        if chain_type in nucleic3to1:
                            newobj_chain = ChainOfNucleotide(pdb = key ,chain = obj_chain.chain)
                            pdbobject._has_nucl = True
                            newobj_chain.fuse(chain = obj_chain, lapl = True)
                            del(obj_chain)
                            obj_chain = newobj_chain
                            pdbobject._chains[pdbobject._get_chain_position_by_id(id = obj_chain.chain)] = obj_chain
                        if chain_type in aminoacids3to1:
                            newobj_chain = ChainOfProtein(pdb = key ,chain = obj_chain.chain)
                            pdbobject._has_prot = True
                            newobj_chain.fuse(chain = obj_chain, lapl = True)
                            del(obj_chain)
                            obj_chain = newobj_chain
                            pdbobject._chains[pdbobject._get_chain_position_by_id(id = obj_chain.chain)] = obj_chain
                    read_translated_line(obj_chain, pdbline)
                    if not pdbobject.chain_exists(chain):
                        pdbobject.add_chain(obj_chain)
                    
            pdbobject._chain_id = set([c.chain for c in pdbobject.chains if not c.is_empty])
            pdbobject._chains   = [c for c in pdbobject.chains if not c.is_empty]

def read_translated_line(thischain, line, keep_version = "A"):
    """
    Given a PDB-formated line, creates an atom to add to a new or a pre-existent residue

    @type  line: String
    @param line: PDB formated line

    @type  keep_version: String
    @param keep_version: Some residues have two versions, codified in front of the residue_type (line[16:17])
                         By default we keep the A version of doubles Aa, but it can be changed through parameters
    """
    isOK = set([" ", keep_version, thischain._residue_version])

    #This skips several versions of the same residue. It always will grab " " and keep_version
    #but in chains that are all B (1EA4_Z) this will help grab them
    if line[16:17] not in isOK:
        if len(thischain) == 0:
            thischain._residue_version = line[16:17]
        else:
            return

    residue_num = int(line[22:26].strip())
    residue_ver = line[26:27]

    #Check the need of creating a new residue
    if thischain.is_empty or (len(thischain._last_appended_list) > 0 and thischain._last_appended_list[-1].identifier != str(residue_num) + residue_ver):
        residue = thischain._new_Residue(number = residue_num, version = residue_ver, Rtype = line[17:20].strip(), mode = line[:6].strip())
        thischain.add_residue(residue)


    #Create the new atom
    x, y, z = [float(line[30+8*i:38+8*i]) for i in range(3)]
    try:
        occupancy = float(line[54:60])
    except:
        occupancy = ""
    try:
        tempFactor = float(line[60:66])
    except:
        tempFactor = ""
    try:
        element = line[76:78].strip()
    except:
        element = ""
    try:
        charge = line[78:80].strip()
    except:
        charge = ""
    atom = thischain._new_Atom(number = line[6:12].strip(), name = line[12:16].strip(), x = x, y = y, z = z, occupancy = occupancy, tempFactor = tempFactor, element = element, charge = charge)

    #Add the atom to the last residue (it applies the same if the residue has been created now or before)
    thischain._last_appended_list[-1].add_atom(atom)
    















# -*-
#
# @author: jaumebonet
# @email:  jaume.bonet@gmail.com
# @url:    jaumebonet.github.io
#
# @date:   2016-02-15 11:08:26
#
# @last modified by:   jaumebonet
# @last modified time: 2016-02-15 14:55:08
#
# -*-
from pynion import Multiton


class SetDict(dict):
    __metaclass__ = Multiton

    def __init__(self, name):
        self._representative = {}

    def __setitem__(self, key, value):
        for k in range(len(key)):
            if k == 0:
                self._representative[value] = key[k]
            super(SetDict, self).__setitem__(key[k], value)

    def representative(self, key):
        v = super(SetDict, self).__getitem__(key)
        return self._representative[v]


Aa_sdict = SetDict('AminoAcids')
a = ['ALA', 'AZT', 'CHA', 'HPH', 'NAL', 'AIB', 'BAL', 'DHA', 'BB9', 'ALM', 'AYA', 'BNN',
     'CHG', 'CSD', 'DAL', 'DNP', 'FLA', 'HAC', 'MAA', 'PRR', 'TIH', 'TPQ', 'BB9']
c = ['CYS', 'CYD', 'CYO', 'HCY', 'CSX', 'SMC', 'BCS', 'BUC', 'C5C', 'C6C', 'CCS', 'CEA',
     'CME', 'CSO', 'CSP', 'CSS', 'CSW', 'CY1', 'CY3', 'CYG', 'CYM', 'CYQ', 'DCY', 'OCS',
     'SOC', 'EFC', 'PR3', 'SCH', 'SCS', 'SCY', 'SHC', 'PEC']
d = ['ASP', 'ASZ', '2AS', 'ASA', 'ASB', 'ASK', 'ASL', 'ASQ', 'BHD', 'DAS', 'DSP']
r = ['ARG', 'ORN', 'ACL', 'ARM', 'AGM', 'HAR', 'HMR', 'DAR']
e = ['GLU', 'GLA', 'GLZ', 'PCA', '5HP', 'CGU', 'DGL', 'GGL', 'GMA']
f = ['PHE', 'DAH', 'HPQ', 'DPN', 'PHI', 'PHL']
g = ['GLY', 'GL3', 'GLZ', 'GSC', 'SAR', 'MPQ', 'NMC', 'MSA', 'DBU']
h = ['HIS', 'HSE', 'HSD', 'HI0', 'HIP', 'HID', 'HIE', '3AH', 'MHS', 'DHI', 'HIC', 'NEP',
     'NEM']
i = ['ILE', 'IIL', 'DIL']
k = ['LYS', 'LYZ', 'ALY', 'TRG', 'SHR', 'LYM', 'LLY', 'KCX', 'LLP', 'DLY', 'DM0']
l = ['LEU', 'NLE', 'LOV', 'NLN', 'NLP', 'MLE', 'BUG', 'CLE', 'DLE', 'MLU']
m = ['MET', 'MSE', 'CXM', 'FME', 'OMT']
n = ['ASN', 'MEN']
p = ['PRO', 'HYP', 'DPR', 'ECQ', 'POM', 'H5M']
q = ['GLN', 'DGN']
s = ['SER', 'HSE', 'STA', 'SVA', 'SAC', 'SEL', 'SEP', 'SET', 'OAS', 'DSN', 'MIS']
t = ['THR', 'PTH', 'ALO', 'TPO', 'BMT', 'DTH', 'CTH']
v = ['VAL', 'NVA', 'DVA', 'DIV', 'MVA']
w = ['TRP', 'TPL', 'TRO', 'DTR', 'HTR', 'LTR']
x = ['ACE', '3FG', 'UNK']
y = ['TYR', 'TYQ', 'TYS', 'TYY', 'TYB', 'STY', 'PTR', 'PAQ', 'DTY', 'IYR', 'GHP', 'D3P',
     'D4P', 'OMZ', 'OMY']
Aa_sdict[a]       = 'A'
Aa_sdict[['ASX']] = 'B'
Aa_sdict[c]       = 'C'
Aa_sdict[d]       = 'D'
Aa_sdict[e]       = 'E'
Aa_sdict[f]       = 'F'
Aa_sdict[g]       = 'G'
Aa_sdict[h]       = 'H'
Aa_sdict[i]       = 'I'
Aa_sdict[['XLE']] = 'J'
Aa_sdict[k]       = 'K'
Aa_sdict[l]       = 'L'
Aa_sdict[m]       = 'M'
Aa_sdict[n]       = 'N'
Aa_sdict[['PYL']] = 'O'
Aa_sdict[p]       = 'P'
Aa_sdict[q]       = 'Q'
Aa_sdict[r]       = 'R'
Aa_sdict[s]       = 'S'
Aa_sdict[t]       = 'T'
Aa_sdict[['SEC']] = 'U'
Aa_sdict[v]       = 'V'
Aa_sdict[w]       = 'W'
Aa_sdict[x]       = 'X'
Aa_sdict[y]       = 'Y'
Aa_sdict[['GLX']] = 'Z'



from collections import deque
import os


class SimilarityMatrix(object):
    '''
    Convert a biomatrix as downloaded from
    ftp://ftp.ncbi.nih.gov/blast/matrices
    to a python dictionary
    '''

    @staticmethod
    def get_matrix(matrixID):
        matdir = os.path.join(os.path.normpath(os.path.dirname(__file__)),
                              'matrices')
        matfile = os.path.join(matdir, matrixID.upper())
        if not os.path.isfile(matfile):
            raise "NO!"  # TODO

        return SimilarityMatrix._parse_matrix(matfile)

    @staticmethod
    def _parse_matrix(matrix_file):
        data   = {}
        aalist = []
        fd = open(matrix_file)
        for line in fd:
            if line.startswith(' '):
                l = line.strip().split()
                for aa in l:
                    data.setdefault(aa, {})
                    aalist.append(aa)
            else:
                l  = deque(line.strip().split())
                aa = l.popleft()
                for i in range(len(aalist)):
                    data[aalist[i]][aa] = int(l[i])

        fd.close()
        return data



# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>

.. affiliation::
    Structural BioInformatics Lab <sbi.upf.edu>
    Baldo Oliva <baldo.oliva@upf.edu>

.. class:: AtomSerie
.. class:: VirtualAtom
"""

# Standard Libraries

# External Libraries
import numpy as np
from numpy.linalg import norm
import pandas as pd

# This Library
from SBILib.structure.io import mmNone
from SBILib.structure.Frame3D import mandatory_fields, frame3D_classify

__all__ = ['AtomSeries', 'VirtualAtom']


class AtomSeries( pd.Series ):
    """A single Atom data.
    """
    _subtyp = 'atom_series'
    _atmtyp = 'ATOM'

    def __init__( self, *args, **kwargs ):
        fields = kwargs.pop('empty_fields', mandatory_fields)
        super(AtomSeries, self).__init__(*args, **kwargs)
        if self.empty:
            data = {}
            for f in fields:
                data.setdefault(f, np.nan)
            super(AtomSeries, self).__init__(data)

        if not set(mandatory_fields).issubset(self.index):
            raise AtomSeriesCreationError('AtomSeries requires a given set of index to be present: {}'.format(','.join(mandatory_fields)))

    # Attributes
    @property
    def coordinates( self ):
        """Spatial position of the :class:`.AtomSeries` as coordinates ``X``, ``Y``, ``Z``.

        :return: :class:`~numpy.ndarray`
        """
        return self[['Cartn_x', 'Cartn_y', 'Cartn_z']].values

    # Boolean Checks
    @property
    def is_empty( self ):
        return (self.isna()).all()

    # Methods
    def distance( self, atom=None ):
        """
        Euclidean distance between two atoms
        """
        if atom is None:
            return norm(self.coordinates - np.zeros(3))
        else:
            return norm(self.coordinates - atom.coordinates)

    def simple_atom( self, atom_id, coordinates ):
        """
        """
        data = mmNone().read()['empty']['_atom_site']

        for i, xyz in enumerate(['x', 'y', 'z']):
            data['Cartn_{}'.format(xyz)] = coordinates[i]
        data['group_PDB'] = self._atmtyp
        for x in ['id', 'label_entity_id', 'label_seq_id', 'auth_seq_id', 'pdbx_PDB_model_num']:
            data[x] = 1
        for x in ['auth_comp_id', 'label_comp_id']:
            data[x] = 'UNK'
        for x in ['auth_asym_id', 'label_asym_id']:
            data[x] = 'A'
        for x in ['auth_atom_id', 'label_atom_id']:
            data[x] = atom_id
        for x in ['occupancy', 'B_iso_or_equiv']:
            data[x] = 1.0
        data['type_symbol'] = data['auth_atom_id'][0] if data['auth_atom_id'] != data['auth_comp_id'] else data['auth_atom_id']

        for k in data:
            if isinstance(data[k], list):
                data[k] = ' '

        return AtomSeries(data)

    # Pandas Magic Methods
    @property
    def _constructor( self ):
        # return AtomSeries
        def f(*args, **kwargs):
            try:
                return frame3D_classify(AtomSeries(*args, **kwargs))
            except AtomSeriesCreationError:
                return frame3D_classify(pd.Series(*args, **kwargs))
        return f

    @property
    def _constructor_expanddim( self ):
        from SBILib.structure import Frame3D

        def f(*args, **kwargs):
            return frame3D_classify(Frame3D(*args, **kwargs))
        return f


class VirtualAtom( AtomSeries ):
    """A single Atom data.
    """
    _subtyp = 'virtual_atom_series'
    _atmtyp = 'VIRATOM'

    def __init__( self, *args, **kw ):
        super(VirtualAtom, self).__init__(*args, **kw)

    @property
    def _constructor( self ):
        return VirtualAtom


class AtomSeriesCreationError( Exception ):
    """Raises an error when a :class:`.AtomSeries` cannot be created.
    """


"""
Sequence

author: jbonet
date:   05/2013


@oliva's lab
"""

import re, copy
from collections import Counter

global AVAILABLE_FORMATS
AVAILABLE_FORMATS = set(['TAB', 'FASTA'])

global gapdefinition
gapdefinition = '[-x]'

class Sequence(object):

	def __init__(self, seqID = '', sequence = ''):
		global gapdefinition

		if bool(re.search(r'\s',seqID)):
			self._seqID = seqID.split()[0]
			self._info  = " ".join(seqID.split()[1:])
		else:
			self._seqID = seqID
			self._info  = ''

		self._sequence = sequence
		self._gapped   = bool(re.search(gapdefinition,self._sequence))

	'''ATTRIBUTES'''
	@property
	def id(self): 			return self._seqID
	@id.setter
	def id(self,value): 	  
		if bool(re.search(r'\s',value)):
			self._seqID = value.split()[0]
			self._info  = " ".join(value.split()[1:])
		else:
			self._seqID = value

	@property
	def sequence(self):	return self._sequence
	@sequence.setter
	def sequence(self, value): 
		self._sequence = value
		self._gapped   = bool(re.search('-',self._sequence))

	@property
	def info(self):		return self._info
	@info.setter
	def info(self, value):
		self._info = value

	@property
	def is_gapped(self):return self._gapped

	'''FUNCTIONS'''
	def contains(self, sequence):
		if isinstance(sequence, str):
			return bool(re.search(sequence, self.sequence))
		elif isinstance(sequence, Sequence):
			return bool(re.search(sequence.sequence, self.sequence))
		else:
			return NotImplemented

	def contained(self, sequence):
		if isinstance(sequence, str):
			return bool(re.search(self.sequence, sequence))
		elif isinstance(sequence, Sequence):
			return bool(re.search(self.sequence, sequence.sequence))
		else:
			return NotImplemented

	def format(self, format = 'TAB'):
		global AVAILABLE_FORMATS
		if format.upper() not in AVAILABLE_FORMATS:
			raise AttributeError('format option not available')

		if format.upper() == 'TAB':
			return '{0.id}\t{0.sequence}'.format(self)
		elif format.upper() == 'FASTA':
			return '>{0.id}\n{0.sequence}'.format(self)

	def tokenize(self, token_coding = None):
		if token_coding is None:
			return copy.deepcopy(self._sequence)
		if token_coding == 'binary':
			tc = ['\D','1','-','0']
			return re.sub(tc[0],tc[1],re.sub(tc[2],tc[3],self._sequence))

	def aa_frequency(self):
		global gapdefinition
		return dict([(x,float(y)/len(self)) for x,y in Counter(re.sub(gapdefinition,'',self._sequence)).items()])

	def duplicate(self):
		return copy.deepcopy(self)

	def do_ungap(self):
		global gapdefinition
		if self._gapped:
			self._sequence = re.sub(gapdefinition,'',self._sequence)

	'''DEFAULT FUNCTIONS OVERWRITTE'''
	def append(self, sequence):
		if isinstance(sequence, str):
			self._sequence += sequence
			if not self._gapped:
				self._gapped = bool(re.search('-',self._sequence))
		elif isinstance(sequence, (tuple, list)):
			self._sequence += "".join(sequence)
			if not self._gapped:
				self._gapped = bool(re.search('-',self._sequence))
		else:
			raise AttributeError()

	def __len__(self):
		return len(self._sequence)

	def __eq__(self, other):
		if isinstance(other, Sequence):
			return self.sequence == other.sequence
		return NotImplemented

	def __ne__(self, other):
		result = self.__eq__(other)
		if result is  NotImplemented:
			return result
		return not result

	def __lt__(self, other):
		if isinstance(other, Sequence):
			return len(self) < len(other)
		return NotImplemented

	def __gt__(self, other):
		if isinstance(other, Sequence):
			return len(self) > len(other)
		return NotImplemented

	def __getitem__(self, key):
		try: 		
			int(key)
			return self._sequence[int(key)]
		except:
			if not isinstance(key, slice): 	
				raise TypeError
			else:
				return self._sequence[key]

	def __iter__(self):
		for s in self._sequence:
			yield s

	def __repr__(self):
		return '<{0.__class__.__name__}: [{0.id}, {0.sequence}]>'.format(self)

	def __str__(self):
		return '{0.id}\t{0.sequence}'.format(self)



"""PDBlink

author: jbonet
date:   10/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
import sys, os, copy
import subprocess
import warnings
import urllib.request, urllib.parse, urllib.error
import ftplib

"""
Dependences in SBI library
"""
from SBILib               import SBIglobals
from SBILib.structure import PDB
from SBILib.beans.Path     import Path
from SBILib.beans.File     import File
from SBILib.databases import PDBftp, PDBrsync
from SBILib.sequence  import Fasta


class PDBlink(object):
    """The PBDlink class controls the management of PDB on database level

        For some functionalities, it requires rsync (in the path). Thus, it might be limited to unix-derived OS.

    """
    def __init__(self, local = None, PDBseq = None):
        self._local    = None if local  is None else os.path.abspath(local)
        self._PDBseq   = None if PDBseq is None else os.path.abspath(PDBseq)
        self.__name__  = 'databases.PDBlink'    # This must be included in every class for the SBIglobals.alert()

    """ATTRIBUTES"""
    @property
    def local(self):         return self._local
    @local.setter
    def local(self, value):  self._local = os.path.abspath(value)

    @property
    def PDBseq(self):        return self._PDBseq
    @PDBseq.setter
    def PDBseq(self, value): self._PDBseq = os.path.abspath(value)

    @property
    def localPDBs(self):
        for pdb_file in Path.list_files(root = self.local, pattern = '*.ent.gz'):
            yield pdb_file

    @property
    def source(self):
        return PDBftp['show']

    """BOOLEANS"""
    @property
    def has_local(self):    return self._local is not None

    """METHODS"""
    def get_PDBseq_filtered(self, resolution_threshold, output_file):
        resolutions     = self.get_resolutions()
        names           = [k for k, v in resolutions.items() if float(v) <= float(resolution_threshold)]
        sequences       = Fasta(os.path.join(self.PDBseq,'PDBseq.fa'))
        selectedseq     = sequences.retrieve(copy.deepcopy(names), prefix_size = 4)
        return Fasta.build_multifasta(output_file, selectedseq, True)

        # outdir   = self.PDBseq if self.PDBseq is not None else os.curdir
        # return os.path.join(outdir, 'PDBseq.{0}.fa'.format(str(resolution_threshold).replace('.','_')))

    def sync_PDB(self, log_file = None):
        if not self.has_local:
            raise NameError('A local PDB database must be defined to sync with.')

        Path.mkdir(self.local)

        command = ['rsync', '-rlpt', '-v', '-z', '--port=' + PDBrsync['port'], PDBrsync['address'],  self.local]

        p = subprocess.Popen(command,
                             stdout = open(log_file,'w') if log_file is not None else subprocess.PIPE,
                             stderr = subprocess.PIPE)

        SBIglobals.alert('verbose', self, 'Executing: {0}'.format(" ".join(command)))

        out, err = p.communicate()
        if err.strip() != '':
            raise SystemError('{0}'.format(err))

    def make_PDBseq(self, log_file, resolution_threshold = None):
        if not self.has_local:
            raise NameError('A local PDB database must be defined to do create a PDBseq database.')
        outdir = self.PDBseq if self.PDBseq is not None else os.curdir

        Path.mkdir(self.PDBseq)
        fasta_file = File(file_name = os.path.join(outdir, 'PDBseq.fa'),     action = 'w', overwrite = True)
        fasta_fd   = fasta_file.descriptor
        idx_file   = File(file_name = os.path.join(outdir, 'PDBseq.fa.idx'), action = 'w', overwrite = True)
        idx_fd     = idx_file.descriptor
        # if resolution_threshold is not None:
        #     filtered_file_name = self.get_PDBseq_filtered(resolution_threshold)
        #     filtered_file      = File(file_name = filtered_file_name, action = 'w', overwrite = True)
        #     filtered_fd        = filtered_file.descriptor
        #     resolutions        = self.get_resolutions(resolution_threshold = resolution_threshold)
        log_file   = File(file_name = log_file, action = 'w', overwrite = True)
        log_idx    = log_file.descriptor

        for pdb_file in self.localPDBs:
            log_idx.write("Reading File: {0}\n".format(pdb_file))
            newPDB = PDB(pdb_file = pdb_file, dehydrate = True)
            fasta_idx = newPDB.FASTA_IDX(nucleotide=False)
            if len(fasta_idx['FASTA']) != len(fasta_idx['IDX']):
                log_idx.write('ERROR!!!!! Number of fastas and indexes are different for pdb {0}!!\n'.format(newPDB.id))
            if len(fasta_idx['FASTA']) > 0:
                log_idx.write('\tPrinting FASTA and IDX...\n')
            else:
                log_idx.write('\tProblably just a nucleotide PDB...\n')
            for c in range(len(fasta_idx['FASTA'])):
                sequence = fasta_idx['FASTA'][c].split('\n')[1]
                sequence = sequence.replace('X','').replace('x','')
                if len(sequence) > 0:
                    fasta_fd.write(fasta_idx['FASTA'][c] + "\n")
                    if resolution_threshold is not None and newPDB.id in resolutions and not newPDB.is_all_ca:
                        filtered_fd.write(fasta_idx['FASTA'][c] + "\n")
                    idx_fd.write(fasta_idx['IDX'][c] + "\n")
            del(newPDB)

        #CLOSE & END
        fasta_file.close()
        idx_file.close()
        if resolution_threshold is not None:
            filtered_fd.close()

    def get_resolutions(self):
        # resolutions (-1) are for methods that do not define resolution
        resolutions = {}

        ftp = ftplib.FTP(PDBftp['address'])
        ftp.login()
        ftp.cwd(PDBftp['derived'])
        resoluIDX = []
        ftp.retrlines('RETR ' + PDBftp['resolution'], resoluIDX.append)
        ftp.quit()

        SBIglobals.alert('debug', self, 'Retrieving resolution data from PDB FTP...')

        active = False
        for line in resoluIDX:
            if line.startswith('-'):
                active = True
                continue
            if active and len(line.strip()) > 0:
                data = [x.strip() for x in line.split(';')]
                if len(data[1]) > 0:
                    SBIglobals.alert('debug', self, '\tResolution for {0[0]} is {0[1]}...'.format(data))
                    # if resolution_threshold is None:
                    resolutions[data[0]] = data[1]

        #rsync is accumulative, we might have structures that are not in the residu.idx anymore.. must check
        for pdb_file in self.localPDBs:
            newfile = File(file_name = pdb_file, action = 'r')
            pdbid   = newfile.prefix.lstrip('pdb').upper()
            if pdbid not in resolutions:
                pdbobj = PDB(pdb_file = pdb_file, header = True, onlyheader = True)
                SBIglobals.alert('debug', self, '\tGrabbing Resolution for {0} is {1}...'.format(pdbid, pdbobj.header.resolution))
                resolutions[pdbid] = pdbobj.header.resolution

        return resolutions

    def get_PDB(self, pdbID):
        if self.has_local:
            rootdir = os.path.join(self.local,pdbID.lower()[1:3])
            for pdb_file in Path.list_files(root = rootdir, pattern = '*.ent.gz'):
                newfile = File(file_name = pdb_file, action = 'r')
                if newfile.prefix.lstrip('pdb').upper() == pdbID.upper():
                    return pdb_file

        #If we do not find it in local (or we do not have a local) we search it on the FTP
        pdb_file = 'pdb' + pdbID.lower() + '.ent.gz'
        source = 'ftp://' + PDBftp['address'] + os.path.join(PDBftp['structures'], pdbID[1:3].lower(), pdb_file)
        try:
            urllib.request.urlretrieve(source, pdb_file)
        except:
            return False
        return os.path.abspath(pdb_file)

    def get_PDBs(self, pdbIDset):
        if isintance(pdbIDset, str):
            warnings.warn('For single PDB search the get_PDB function is recomended.')
            yield self.get_PDB(pdbIDset)
        else:
            pdbIDset = set([x.upper() for x in pdbIDset])

        if self.has_local:
            for pdb_file in self.localPDBs:
                newfile = File(file_name = pdb_file, action = 'r')
                if newfile.prefix.lstrip('pdb').upper() in pdbIDset:
                    yield pdb_file
        else:
            for pdbID in pdbIDset:
                yield self.get_PDB(pdbID)

    def get_FASTA_IDX_by_names_to_file(self, names, outfile):


        fastafile     = Fasta(self.PDBseq)
        selectedfasta = fastafile.retrieve(copy.deepcopy(names))
        output_fasta  = File(outfile,'w')
        for sequence in selectedfasta:
            output_fasta.write(sequence.format('FASTA') + "\n")
        output_fasta.close()
        idxfile       = self.PDBseq + '.idx'
        output_idx    = File(outfile + '.idx','w')
        input_idx     = File(idxfile,'r')
        for line in input_idx.descriptor:
            info    = line.split()
            pdbname = info[0][1:]
            if pdbname in names:
                output_idx.write(line)
        input_idx.close()
        output_idx.close()







'''
jbonet @ oliva's lab 2013
'''
import re

from SBILib.error    import SeqAliError as SAE
from SBILib.sequence import SeqAli      as SeqAli
from SBILib.beans.IndexedNum    import IndexedNum

class IndexedSeqAli(SeqAli):

	def __init__(self, sequences, sequenceInits, identities = None, positives = None, gaps = None):

		super(IndexedSeqAli, self).__init__(sequences, sequenceInits, identities, positives, gaps)

		self._has_complex_index = self._check_index_kind(sequenceInits)

	'''FUNCTIONS CALLED ON LOAD'''
	def _check_index_kind(self, index):
		for idx in index:
			if isinstance(idx, list): return True
		return False

	'''INTERNAL GETTERS/SETTERS'''
	def _sequence_position_id(self, refseq, pos): #Only for ungapped (basically to use in _search_segments())
		if not isinstance(self._idx[refseq], list):
			return super(IndexedSeqAli, self)._sequence_position_id(refseq, pos)
		else:
			return self._idx[refseq][pos - 1]

	def _sequence_position_from_alignment_position(self, refseq, pos):
		#Considering over alignment with everything
		#if the alignemnt position is gap, we move up to the next position
		if not isinstance(self._idx[refseq], list):
			return super(IndexedSeqAli, self)._sequence_position_from_alignment_position(refseq, pos)
		else:

			data = self._seq[refseq][:pos]
			c    = 1
			while (data.endswith('-') and (data <= len(self._seq[refseq]))):
				data = self._seq[refseq][:pos + c]
				c += 1
			return self._idx[refseq][len(data) - 1]

	def _alignment_position_from_sequence_position(self, refseq, pos):
		if not isinstance(self._idx[refseq], list):
			return super(IndexedSeqAli, self)._alignment_position_from_sequence_position(refseq, pos)

		pos = IndexedNum(pos)
		if not self._seq[refseq].is_gapped:
			try: return self._idx[refseq].index(pos) + 1
			except ValueError as e: raise IndexError(e.message)
		else:
			c = 0
			pregaps = True
			for x in range(len(self._seq[refseq])):
				if pregaps and self._seq[refseq][x] == '-': continue
				pregaps = False
				if pos == self._idx[refseq][c]:
					return x + 1
				if x+1 < len(self._seq[refseq]) and self._seq[refseq][x+1] != '-':
					c += 1
			raise IndexError

	'''FUNCTIONS'''
	def add_complex_index(self, refseq, index, pregaps = 0):
		if isinstance(self._idx[refseq], list):
			raise AttributeError('A complex index cannot be modified')
		if not isinstance(index, list):
			raise AttributeError('A complex index is a list of IndexedNum or integers')
		if len(index) < len(re.sub('-','',self._seq[refseq].sequence)):
			raise AttributeError('The given index is not long enought to cover que assigned sequence')

		init_value  = self._idx[refseq]
		part_to_add = index[init_value - 1 - pregaps:]
		self._idx[refseq] = []
		count             = 0

		for x in range(len(self._seq[refseq])):
			if self._seq[refseq][x] != '-':
				self._idx[refseq].append(IndexedNum(part_to_add[count]))
				count += 1

		for x in range(len(self._segment[refseq])):
			if self._segment[refseq][x] != '-':
				self._segment[refseq][x] = self._sequence_position_id(refseq, self._segment[refseq][x] - init_value + 1)

		self._has_complex_index = True

	'''OVEWRITTE OBJECT FUNCTIONS (FROM PARENT)'''
	def __getitem__(self, key):
		# We correct the count as one would try to ask for the first position as 1
		# Runs by alignment position
		# if not self._has_complex_index:
		# 	return super(IndexedSeqAli, self).__getitem__(key)
		try:
			int(key)
			if int(key) > len(self): raise IndexError
			return [n[int(key) - 1] for n in self._seq]
		except:
			if not isinstance(key, slice):
				raise TypeError()
			else:
				index = key.indices(len(self) + 1)
				if key.stop is not None:
					index = (index[0],index[1]+1,index[2]) #Remember! we go according to align position, not array!!
				else:
					index = (index[0],index[1],index[2])
				sequences = ['' for x in range(self.number_of_sequences)]
				seqInits  = ['' for x in range(self.number_of_sequences)]
				complex_indexes = {}
				complex_pregaps = {}
				for i in range(*index):
					data = self[i]
					for x in range(len(data)):
						sequences[x] += data[x]
					if i == len(self): break

				for x in range(self.number_of_sequences):
					if not isinstance(self._idx[x], list):
						seqInits[x] = self._sequence_position_from_alignment_position(x,index[0])
					else:
						seqInits[x] = self._idx[x].index(self._sequence_position_from_alignment_position(x,index[0])) + 1
						complex_indexes[x] = self._idx[x]
						complex_pregaps[x] = Counter(str(self.sequences[x]).split(str(sequences[x]))[0])['-']
				newali = self.__class__(sequences, seqInits)
				for refident in complex_indexes:
					newali.add_complex_index(refident, complex_indexes[refident], complex_pregaps[refident])
				if not self._alipatt is None:
					newali.add_alignment_pattern(self._alipatt[index[0]-1:index[1]-1:index[2]],self._aliptmeth)
				return newali



"""
Chain

author: jbonet
date:   02/2013

@oliva's lab
"""

import warnings
import copy
import numpy
import re

from ..atom       import Atom
from ..residue    import Residue, ResidueOfAminoAcid, ResidueOfNucleotide
from SBILib.data     import aminoacids_main1
from SBILib          import SBIglobals


class Chain(object):
    """
    A {Chain} represents a collection of {Residue}s
    """
    chaintype   = 'O'
    atomtype    = Atom
    resitype    = Residue
    dictitype   = {}

    def __init__(self, pdb, chain):
        """
        @type  pdb: String
        @param pdb: PDB identifier

        @type  chain: String
        @param chain: Chain identifier
        """

        if chain == "":
            #warnings.warn("Possible old format PDB. No-chain is transformed to chain 'A'")
            chain = "A"

        # IDENTIFIERS
        self._pdb   = re.sub(r'^PDB','',pdb.upper())
        self._chain = chain
        self._altID = None

        # LISTS
        self._structure = []
        self._hetero    = []
        self._HOH       = []

        self._chaindict = {}

        # PROPERTIES
        self._atom_length = None

        # RESIDUE VERSION ON READ
        self._residue_version = None

        # TERM
        self._term = False

        # CONTROL
        self._last_appended_list = []
        self._numeric_version    = 0 #pdb: 1CJ0

    #
    # ATTRIBUTES
    #
    @property
    def globalID(self):
        """
        {Chain} global identifier
        @rtype: String
        """
        return "{0}_{1}".format(self._pdb, self._chain)

    @property
    def pdb(self):
        """
        PDB name
        @rtype: String (uppercase)
        """
        return self._pdb

    @property
    def chain(self):
        """
        Chain identifier
        @rtype: String
        """
        return self._chain

    @chain.setter
    def chain(self, value):
        """
        Changes the chain assigned to the {Chain}
        @type value: String
        """
        self._chain = value

    @property
    def waters(self):
        """
        Lists, if any, the waters in the {Chain}
        @rtype: List of {Atom}s with type HOH
        """
        return self._HOH

    @property
    def heteroatoms(self):
        """
        List, if any, of NOT AMINOACID/NUCLEOTIDE heteroatoms
        @rtype: List of {Atom}s
        """
        return self._hetero

    @property
    def full_chain(self):
        return self._structure + self._hetero + self._HOH

    @property
    def alternative_id(self):
        """
        Returns an alternative ID that can be added in some processes
        @rtype: String
        """
        return self._altID

    @alternative_id.setter
    def alternative_id(self, value):
        """
        Sets an alternative ID
        @typr value: String
        """
        self._altID = value

    @property
    def atom_length(self):
        """
        Total number of atoms of the {Chain}
        @rtype: Integer
        """
        if self._atom_length != None:
            return self._atom_length
        else:
            self._atom_length = sum(map(len,self._all_residues))
            return self._atom_length

    @property
    def last_residue(self):
        """
        Returns the GLOBAL last residue
        @rtype: {Residue}
        """
        return self._all_residues[-1]

    @property
    def first_structure(self):
        return self._structure[0] if len(self._structure) > 0 else self._hetero[0]
    @property
    def last_structure(self):
        return self._structure[-1] if len(self._structure) > 0 else self._hetero[-1]

    @property
    def is_term(self):
        """
        Mark that we have read untill chain TERM, from here everything will count as heteroatoms
        (only for original PDBs)
        """
        self._term = True

    #
    # COMPLEX GETTERS & SETTERS
    #
    def get_residue_by_identifier(self, identifier):
        if len(self._chaindict) == 0:
            for r in self.full_chain:
                self._chaindict[r.identifier.strip()] = r
        try:
            r = self._chaindict[str(identifier.strip())]
        except KeyError as e:
            SBIglobals.alert('error', self, '{0} is not a valid identifier for chain {1}'.format(identifier, self.globalID), e)
        else:
            return r

    def add_residue(self, residue):
        """
        Appends a new {Residue} to the list

        @type residue: {Residue}
        """
        if isinstance(residue, ResidueOfAminoAcid) or isinstance(residue, ResidueOfNucleotide):
            self._structure.append(residue)
            self._last_appended_list = self._structure
        elif residue.type == 'HOH' or residue.type == 'DOD':
            self._HOH.append(residue)
            self._last_appended_list = self._HOH
        else:
            self._hetero.append(residue)
            self._last_appended_list = self._hetero

    def join(self, chain):
        self._hetero.extend(chain._hetero)
        self._structure.extend(chain._structure)
        self._HOH.extend(chain._HOH)

    #
    # PRIVATE ATTRIBUTES
    #
    @property
    def _all_residues(self):
        """
        Returns all the 3 types of residues
        @rtype: List of {Residue}s
        """
        return self._structure + self._hetero + self._HOH

    def _all_atoms_coordinates(self, structure = True, hetero = True, water = True, geometric_center = False):
        all_coord = numpy.array(numpy.zeros(3))
        res_array = []
        if structure: res_array = res_array + self._structure
        if hetero:    res_array = res_array + self._hetero
        if water:     res_array = res_array + self._HOH
        for residue in res_array:
            if not geometric_center:
                all_coord = numpy.vstack((all_coord, residue.all_coordinates))
            else:
                all_coord = numpy.vstack((all_coord, residue.geometric_center))
        all_coord = numpy.delete(all_coord, 0, 0)
        return all_coord

    def _get_structure_array_coordinate(self, identifier):
        if isinstance(identifier, int):
            identifier = str(identifier + ' ')
        for i in range(len(self._structure)):
            if self._structure[i].identifier == identifier:
                return i
    #
    # METHODS
    #
    def dehydrate(self):
        """
        Removes all the waters
        """
        del(self._HOH)
        self._HOH = []

    def extract(self, init = None, end = None, backbone = False):
        """
        returns a section of a Chain
        """
        new_chain = self.__class__(self.pdb, self.chain)
        try:    init_n, init_i = int(init), ' '
        except: init_n, init_i = int(init[:-1]), init[-1]
        try:    end_n,  end_i  = int(end),  ' '
        except: end_n,  end_i  = int(end[:-1]),  end[-1]

        capture = False
        for residue in self._structure:
            if residue.number == init_n and residue.version == init_i: capture = True
            if capture: new_chain.add_residue(residue)
            if residue.number == end_n and residue.version == end_i:
                capture = False
                break

        nchain = new_chain.duplicate()
        if backbone:
            for residue in nchain.aminoacids:
                residue.remove_side_chain()

        return nchain

    def rotate(self, matrix):
        """
        Rotates each {Residue} according to a given matrix

        @type matrix: numpy.matrix
        """
        for residue in self._all_residues:
            residue.rotate(matrix = matrix)

    def translate(self, vector):
        """
        Translates each {Residue} according to a translational vector

        @type vector: numpy.array
        """
        for residue in self._all_residues:
            residue.translate(vector = vector)

    def reposition(self, matrix = None, vector = None):
        """
        Rotates and Translates each {Residue} according to a matrix and a translational vector

        @type matrix: numpy.matrix

        @type vector: numpy.array
        """
        if matrix is None: matrix = numpy.identity(3, float)
        if vector is None: vector = numpy.zeros(3, float)
        for residue in self._all_residues:
            residue.reposition(matrix = matrix, vector = vector)


    def translate_onto_origin(self):
        self.translate(vector=(self.geometric_center()) * -1)

    def geometric_center(self, structure = True, hetero = True, water = True, by_residue = False):

        all_atoms = self._all_atoms_coordinates(structure        = structure,
                                                hetero           = hetero,
                                                water            = water,
                                                geometric_center = by_residue)
        if by_residue: return all_atoms
        else:          return numpy.sum(all_atoms,axis=0) / all_atoms[:,1].size

    def clean(self, initatom = 1, initresidue = 1, nothetero = True, normalize = True):
        try:
            initatom    = int(initatom)
            initresidue = int(initresidue)
        except: raise AttributeError
        old_number = -99999

        if nothetero: working_array = self._structure
        else:         working_array = self._all_residues
        for x in range(len(working_array)):
            #Atoms
            if initatom > 99999: initatom = 1
            working_array[x].renumerate_atoms(first_atom_number = initatom)
            initatom = working_array[x].last_atom_number + 1
            #Residues
            original_number = working_array[x].number
            if old_number != -99999:
                if old_number != original_number:
                    if original_number > old_number:
                        initresidue += original_number - old_number
                    else:
                        if working_array[x].follows(working_array[x-1]):
                            initresidue += 1
                        else:
                            initresidue += 2
                else:
                    initresidue += 1
            working_array[x].number  = initresidue
            working_array[x].version = ' '
            if normalize and isinstance(working_array[x], ResidueOfAminoAcid):
                if working_array[x].single_letter in aminoacids_main1:
                    working_array[x].normalize()
            old_number = original_number

    def renumerate_residues(self, init = 1, gaps = True):
        """
        Renumerates a chain starting in a given initial number

        @type  init: Integer
        @param init: Starting number of the renumerated chain.
                     1 by default.

        @type  gaps: Boolean
        @param gaps: Defines if the gaps in the chain have to be kept.
                     True by default.

        @raise AttributeError if init is not or cannot be casted to an integer
        """

        try:    init = int(init)
        except: raise AttributeError

        for x in range(len(self)):
            original_number = self._all_residues[x].number
            self._all_residues[x].number = init
            if gaps:
                if x+1 < len(self):
                    init = init + (self._all_residues[x+1].number - original_number)
            else:
                init = init + 1

    def renumerate_atoms(self, init = 1):
        """
        Renumerates de atoms of a chain consecutively from a given value
        @type  init: Integer
        @param init: Starting atom number of the renumerated chain.
                     1 by default.

        @raise AttributeError if init is not or cannot be casted to an integer
        """

        try:    init = int(init)
        except: raise AttributeError

        for x in range(len(self)):
            if init > 99999: init = 1
            self._all_residues[x].renumerate_atoms(first_atom_number = init)
            init = self._all_residues[x].last_atom_number + 1

    def duplicate(self, hetero = True, water = False, backbone = False):
        """
        Returns a {Chain} identical to the original but as a new object
        @rtype: {Chain}
        """
        new_chain = self.__class__(pdb = self._pdb, chain = self._chain)

        if not water:  HOH_chain    = None
        else:          HOH_chain    = copy.deepcopy(self._HOH)
        if not hetero: hetero_chain = None
        else:          hetero_chain = copy.deepcopy(self._hetero)

        new_chain._add_residues(structure = copy.deepcopy(self._structure),
                                hetero    = hetero_chain,
                                HOH       = HOH_chain)
        if backbone:
            for residue in new_chain.aminoacids:
                residue.remove_side_chain()

        return new_chain

    def fuse(self, chain, lapl = False):
        """
        Merges another chain over self.
        @type chain: {Chain}
        """
        self._add_residues(structure = chain._structure,
                           hetero    = chain.heteroatoms,
                           HOH       = chain.waters)
        if lapl:
            self._last_appended_list = chain._last_appended_list
            self._residue_version = chain._residue_version

    #
    # PRIVATE METHODS
    #
    def _add_residues(self, structure = None, hetero = None, HOH = None):
        """
        Appends a list of {Residues} to the residue list

        @type residues: Array of {Residue}
        """
        if structure is not None:
            self._structure.extend(structure)
        if hetero is not None:
            self._hetero.extend(hetero)
        if HOH is not None:
            self._HOH.extend(HOH)

    def _new_Residue(self, number, version, Rtype, mode):
        """
        Creates a new {Residue}

        @type  number: Integer
        @param number: Residue number

        @type  version: Char
        @param version: Optional char used on pdbs to change count

        @type  type: String
        @param type: Residue type

        @type  mode: String
        @param mode: Residue mode: ATOM or HETATM

        @rtype: {Residue}
        """
        if Rtype in self.dictitype and not self._term:
            return self.resitype(number = number, version = version, Rtype = Rtype, mode = mode)
        else:
            return Residue(number = number, version = version, Rtype = Rtype, mode = mode)

    def _new_Atom(self, number, name, x, y, z, occupancy, tempFactor, element, charge):
        """
        Creates a new {Atom}

        @type  number: Integer
        @param number: Atom number

        @type  name: String
        @param name: Atom identifier

        @type  x: Decimal
        @param x: x coordinate

        @type  y: Decimal
        @param y: y coordinate

        @type  z: Decimal
        @param z: z coordinate

        @rtype: {Atom}
        """
        if isinstance(self._last_appended_list[-1], self.resitype):
            return self.atomtype(number = number, name = name, x = x, y = y, z = z, occupancy = occupancy, tempFactor = tempFactor, element = element, charge = charge)
        else:
            return Atom(number = number, name = name, x = x, y = y, z = z, occupancy = occupancy, tempFactor = tempFactor, element = element, charge = charge)

    def _chain_idx(self):
        """
        Returns indexes (number + version) of the residues separated by x on gaps
        @rtype: String
        """
        data = []
        pos = self._structure[0].number - 1
        for r in self._structure:
            if r.number == pos + 1:
                data.append(r.identifier)
            else:
                #this jump size is due to PDBs badly annotated
                if r.number - pos - 1 < 650:
                    for x in range(r.number - pos - 1):
                        data.append('X')
                data.append(r.identifier)
            pos = r.number

        return ";".join(data)

    #
    # BOOLEANS
    #
    def residue_exists(self, identifier):
        """
        Confirms if a given residue exists in the Chain
        @rtype: Boolean
        """
        if len(self._chaindict) == 0:
            for r in self.full_chain:
                self._chaindict[r.identifier.strip()] = r

        return identifier in self._chaindict

    @property
    def has_heteroatoms(self):
        """
        Returns True if there is any NON AMINOACID/NUCLEOTIDE heteroatom. False otherwise
        @rtype: Boolean
        """
        return len(self.heteroatoms) > 0

    @property
    def has_water(self):
        """
        Returns True if there is any HOH heteroatom. False otherwise
        @rtype: Boolean
        """
        return len(self.waters) > 0

    @property
    def is_empty(self):
        """
        Returns True if there isn't any {Residue} in the {Chain}. False otherwise
        @rtype: Boolean
        """
        return not self.has_water and not self.has_heteroatoms and len(self._structure) == 0

    #
    # IO
    #
    def read_PDB_line(self, line, keep_version = "A"):
        from ..parse_pdb import read_PDB_line
        read_PDB_line(self, line, keep_version)

    def PDB_format(self, terminal = True):
        """
        Strings a {Chain} in PDB format
        @rtype: String
        """
        lines = []
        residue = None
        # for residue in self._all_residues:
        for residue in self._structure:
            for atom in residue.atoms:
                if atom.occupancy != "" and atom.tempFactor != "":
                    lines.append("%s%s %s%s%s%s%s%s%s%s%s          %s%s"
                                %(residue.mode.ljust(6), str(atom.number).rjust(5),
                                str(atom.pretty_name),
                                residue.standard_type.rjust(4),
                                str(self._chain).rjust(2),
                                str(residue.identifier).rjust(5),
                                ("%.3f" %float(atom.x)).rjust(11),
                                ("%.3f" %float(atom.y)).rjust(8),
                                ("%.3f" %float(atom.z)).rjust(8),
                                ("%.2f" %float(atom.occupancy)).rjust(6),
                                ("%.2f" %float(atom.tempFactor)).rjust(6),
                                str(atom.element).ljust(2),
                                str(atom.charge).ljust(2)))
                else:
                    lines.append("%s%s %s%s%s%s%s%s%s%s%s          %s%s"
                                %(residue.mode.ljust(6), str(atom.number).rjust(5),
                                str(atom.pretty_name),
                                residue.standard_type.rjust(4),
                                str(self._chain).rjust(2),
                                str(residue.identifier).rjust(5),
                                ("%.3f" %float(atom.x)).rjust(11),
                                ("%.3f" %float(atom.y)).rjust(8),
                                ("%.3f" %float(atom.z)).rjust(8),
                                str(atom.occupancy).rjust(6),
                                str(atom.tempFactor).rjust(6),
                                str(atom.element).ljust(2),
                                str(atom.charge).ljust(2)))
        if terminal:
            if residue is not None:
                lines.append("TER%s%s%s" %(residue.type.rjust(17),str(self._chain).rjust(2),
                                           str(residue.number).rjust(4)))

        return "\n".join(lines)

    def js_format(self, center=False):
        init = 'var {0}="'.format('pdb_' + self.globalID)
        return init + self.PDB_format().replace('\n', '\\n') + '"'

    #
    # OVERRIDE DEFAULT METHODS
    #
    def __len__(self):
        return len(self._all_residues)

    def __repr__(self):
        repre = []
        repre.append('<{0.__class__.__name__}: [{0.pdb}, {0.chain}]>'.format(self))
        for residue in self._all_residues:
            repre.append(repr(residue))
        return "\n".join(repre)



# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>
.. codeaffiliation:: Structural BioInformatics Lab <baldo.oliva@upf.edu>
"""

# Standard Libraries

# External Libraries
import six
import request
import pandas as pd

# This Library
import SBI.core as core

class PDBeChemConnect( object ):
    @staticmethod
    def retrieve_database():
        url = core.get_option('data', 'pdbechem')


import os
import re

from SBILib.databases import DBlink
from uniprot       import Uniprot
from SBILib.beans     import File
from SBILib           import SBIglobals as SBIg


class Connect(DBlink):
    '''
    Manages the connection to the [Uniprot](http://www.uniprot.org/) database.
    '''
    _FTP = 'ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/'
    _MANDATORY_FILES = ['swisprot.gz', 'swisprot.fasta',
                        'trembl.gz',   'trembl.fasta']
    _ITEM_FILES      = ['swisprot.gz', 'trembl.gz']
    _SOURCES         = ['uniprot_sprot.dat.gz', 'uniprot_trembl.dat.gz']
    _SHOW_LINK       = 'http://www.uniprot.org/'

    _DBOBJECT        = Uniprot

    def __init__(self, local):
        '''
        @param:    local
        @pdef:     local directory to store the database. Create if not exist
        @ptype:    {String}
        '''
        super(Connect, self).__init__(local)
        self._RELEASE['db'] = 'uniprot'

    ###################
    # PRIVATE METHODS #
    ###################
    def _process(self, update = False):
        '''
        Transform the source files into the final local db files.

        @param:    update
        @pdef:     toggles between create and update processing
        @pdefault: _False_
        @ptype:    {Boolean
        '''
        if update:
            old = self._RELEASE['total_items'].copy()
        j = 0
        for i in range(len(self._SOURCES)):
            dfilen  = os.path.join(self.local, self._SOURCES[i])
            ofilen  = os.path.join(self.local, self._MANDATORY_FILES[j])
            ffilen  = os.path.join(self.local, self._MANDATORY_FILES[j + 1])
            if not os.path.isfile(dfilen):
                continue
            SBIg.alert('verbose', self, 'Parsing:       {0}'.format(dfilen))
            SBIg.alert('verbose', self, 'DB file to:    {0}'.format(ofilen))
            SBIg.alert('verbose', self, 'Fasta file to: {0}'.format(ffilen))
            dfile   = File(dfilen)
            ofile   = File(ofilen, 'w', update)
            ffile   = File(ffilen, 'w', update)
            protein = None
            for protein in Connect._parse_uniprot(dfile):
                pname = protein.entry_name
                pvers = protein.version
                SBIg.alert('verbose', self, 'Protein: {0}'.format(pname))
                if not update:
                    self._RELEASE['total_items'][pname] = pvers
                else:
                    if pname not in self._RELEASE['total_items']:
                        self._RELEASE['new_items'][pname] = pvers
                    else:
                        del(old[pname])
                        if self._RELEASE['total_items'][pname] != pvers:
                            self._RELEASE['update_items'][pname] = pvers

                ffile.write(protein.sequence.format('FASTA') + '\n')
                ofile.write(protein.json() + '\n')
            j += 2
            dfile.close()
            ofile.close()
            ffile.close()

        if update:
            self._RELEASE['total_items'].update(self._RELEASE['new_items'])
            self._RELEASE['total_items'].update(self._RELEASE['update_items'])
            self._RELEASE['deleted_items'] = old
            for k in self._RELEASE['deleted_items']:
                del(self._RELEASE['total_items'][k])

    @staticmethod
    def _parse_uniprot(uniprot_file):
        '''
        Reads a uniprot file and generates the different uniprot entries

        @param:    uniprot_file
        @pdef:     uniprot file to parse
        @ptype:    {File}

        @yields: {Uniprot}
        '''
        version = re.compile('DT\s+\d+\-\w+\-\d+\,\sentry\sversion\s(\d+)')
        taxid   = re.compile('NCBI_TaxID\=(\d+);')
        protein = None
        for line in uniprot_file.read():
            line = line.rstrip()
            if line.startswith('ID'):
                protein = Uniprot(line.split()[1].strip(),
                                  line.split()[2].strip(' ;'))
            if line.startswith('DT'):
                m = version.search(line)
                if m:
                    protein.version = m.group(1)
            if line.startswith('AC'):
                protein.accession = [x.strip(';') for x in line.split()[1:]]
            if line.startswith('OX'):
                protein.taxid = taxid.search(line.split()[1]).group(1)
            if line.startswith('OH'):
                protein.host = taxid.search(line.split()[1]).group(1)
            if line.startswith('DR'):
                protein.referals = ''.join(line.split()[1:]).split(';')
            if line.startswith('  '):
                protein.sequence = line.replace(' ', '')
            if line.startswith('//'):
                yield protein



"""
Fasta

author: jbonet
date:   05/2013


@oliva's lab
"""

import subprocess, sys

from SBILib.beans.File    import File
from SBILib.sequence import Sequence
from SBILib              import SBIglobals

class Fasta(object):

    def __init__(self, fasta_file):

        if isinstance(fasta_file, str):
            self._file = File(file_name = fasta_file, action = 'r')
        elif isinstance(fasta_file, File):
            self._file = fasta_file
            self._file.action = 'r'
        else:
            raise AttributeError('Check the input of the Fasta object')

        self._is_multifasta = self._check_multifasta()

        self._sequences = []
        self._seqfinder = {}

    '''ATTRIBUTES'''
    @property
    def file(self):          return self._file

    @property
    def is_multifasta(self): return self._is_multifasta

    @property
    def sequences(self):     return self._sequences

    '''FUNCTIONS'''
    def load(self):
        for line in self.file.descriptor:
            if line.startswith('>'):
                self._sequences.append(Sequence(seqID = line.lstrip('>').strip()))
                self._seqfinder[self._sequences[-1].id] = len(self._sequences) - 1
            elif len(line.strip()) > 0:
                self._sequences[-1].append(line.strip())
        self.file.close()

    def live_show(self):
        n = 0
        for line in self.file.descriptor:
            if line.startswith('>'):
                if n > 0:
                    yield s
                n += 1
                s = Sequence(seqID = line.lstrip('>').strip())
            elif len(line.strip()) > 0:
                s.append(line.strip())
        self.file.close()
        yield s

    def retrieve(self, seqID, allbut = False, prefix_size = None):
        SBIglobals.alert('debug', self, 'Getting sequence for {0}'.format(seqID))

        if isinstance(seqID, str):
            if len(self) == 0:
                sequence = Sequence()
                read = False
                for line in self.file.descriptor:
                    if line.startswith('>'):
                        if prefix_size is not None: sid = line.lstrip('>').split()[0].strip()[:prefix_size]
                        else:                       sid = line.lstrip('>').split()[0].strip()
                        if sid == seqID:
                            sequence.id = line.lstrip('>').split()[0].strip()
                            read = True
                        elif read: break
                    elif read and len(line.strip()) > 0:
                        sequence.append(line.strip())
                self.file.close()
            else:
                if not seqID in self._seqfinder:
                    raise KeyError(seqiID)
                return self._sequences[self._seqfinder[seqID]]

        if isinstance(seqID, list): seqID = set(seqID)
        if isinstance(seqID, set):
            sequence = []
            if len(self) == 0:
                read  = False
                for line in self.file.descriptor:
                    if line.startswith('>'):
                        if prefix_size is not None: sid = line.lstrip('>').split()[0].strip()[:prefix_size]
                        else:                       sid = line.lstrip('>').split()[0].strip()
                        if (not allbut and sid in seqID) or \
                           (allbut and sid not in seqID):
                            newSeq = Sequence(seqID = line.lstrip('>').split()[0].strip())
                            sequence.append(newSeq)
                            if not allbut and prefix_size is None:
                                seqID.remove(newSeq.id)
                            read = True
                        elif read: read = False
                        elif len(seqID) == 0: break
                    elif read and len(line.strip()) > 0:
                        sequence[-1].append(line.strip())
                self.file.close()
            else:
                for queryID in seqID:
                    if not queryID in self._seqfinder:
                        raise KeyError(queryID)
                    sequence.append(self._sequences[self._seqfinder[seqID]])

        return sequence

    def get_all_ids(self):
        identifiers = ()
        if len(self) == 0:
            for line in self.file.descriptor:
                if line.startswith('>'):
                    identifiers.append(line.lstrip('>').split()[0].strip())
            self.file.close()
        else:
            for seq in self:
                identifiers.append(seq.id)
        return identifiers

    @staticmethod
    def build(file_name, sequenceID, sequence, force = False):
        newFasta = File(file_name,'w', overwrite = force)
        newSeq   = Sequence(seqID = sequenceID, sequence = sequence)
        file_dsc = newFasta.descriptor
        file_dsc.write(newSeq.format('FASTA'))
        newFasta.close()
        return Fasta(fasta_file = newFasta.full)

    @staticmethod
    def build_multifasta(file_name, sequenceList, force = False):
        newFasta = File(file_name,'w', overwrite = force)
        file_dsc = newFasta.descriptor
        for sequence in sequenceList:
            file_dsc.write(sequence.format('FASTA')+"\n")
        newFasta.close()
        return Fasta(fasta_file = newFasta.full)

    '''PRIVATE FUNCTIONS'''
    def _check_multifasta(self):

        grep_search  = ['grep','-c','>',self.file.full]
        grep_process = subprocess.Popen(grep_search, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        g_out, g_err = grep_process.communicate()

        if int(g_out.strip()) > 1: return True
        else:                      return False

    '''DEFAULT FUNCTIONS OVERWRITTE'''
    def __len__(self):
        return len(self._sequences)

    def __getitem__(self, key):
        try:    int(key)
        except: raise TypeError
        return self._sequences[int(key)]

    def __iter__(self):
        for s in self._sequences:
            yield s

    def __repr__(self):
        return '<{0.__class__.__name__}: [{0.file}]>'.format(self)

    def __str__(self):
        if len(self) == 0:
            return self.file.full
        else:
            text = []
            for seq in self._sequences:
                text.append(seq.format('FASTA'))
            return "\n".join(text)



"""
BlastResult

The BlastResult object contains the data from a single blastpgp XML file.

WARNING!: It has to be a blastpgp from a __single sequence__; multi-fasta blasts are not compatible
          with xml.dom.minidom parsing capabilities.

                                   ############################################
                                        jbonet @ boliva's lab        2011
"""

'''
    Supporting Modules
'''
import os, sys, pydoc, warnings, gzip
from SBILib.error import BlastError as BE
from SBILib.beans.StorableObject import StorableObject
from SBILib           import SBIglobals

try:
    import pickle as pickle
except:
    import pickle

class BlastResult(StorableObject):

    def __init__(self, queryname   = None, querylength     = 0,    blastversion = None,
                       blastmatrix = None, gap_open        = 0,    gap_extend   = 0,
                       blastdb     = 0,    BlastResultFile = None, queryseq     = None):
        """
        The BlastResult object contains the data from a single blastpgp XML file.

        @type  queryname:       String
        @param queryname:       Name of the query sequence.
                                B{MANDATORY} unless I{BlastResultFile} is given.

        @type  querylength:     Integer
        @param querylength:     Length of the query sequence.
                                B{MANDATORY} unless I{BlastResultFile} is given.

        @type  blastversion:    String
        @param blastversion:    Version of blastpgp that created the L{BlastResult}.

        @type  blastmatrix:     String
        @param blastmatrix:     Weight matrix used in the blastpgp run.

        @type  gap_open:        Integer
        @param gap_open:        Score used in the blastpgp run.

        @type  gap_extend:      Integer
        @param gap_extend:      Score used in the blastpgp run.

        @type  blastdb:         String
        @param blastdb:         Sequence database over which the search was performed.

        @type  BlastResultFile: String
        @param BlastResultFile: File containing a L{BlastResult} object to upload.

        @type  queryseq:        String
        @param queryseq:        Query sequence.

        @raise AtributeError: IF mandatory paremeters are not met and BlastResultFile is not given.
        """

        if BlastResultFile != None:
            self.__dict__ = BlastResult.load(object_file = BlastResultFile).__dict__

        else:
            self._query          = queryname         # Query Name
            self._query_length   = int(querylength)  # Query Sequence Length
            self._query_sequence = queryseq          # Query Sequence
            self._version        = blastversion      # Blast version that has been executed for the query
            self._matrix         = blastmatrix       # Similarity matrix used
            self._gap_open       = int(gap_open)     # Gap opening penalty
            self._gap_extend     = int(gap_extend)   # Gap extension penalty
            self._db             = blastdb           # Database used in the search
            self._lastiteration  = 0                 # Keep the last iteration
            self._hits           = []                # List of BlastHit objects

            self._correctedHits  = False
            self._correctFile    = None

    '''ATTRIBUTES'''
    @property
    def query(self):                return self._query
    @property
    def query_length(self):         return int(self._query_length)
    @property
    def query_sequence(self):       return self._query_sequence
    @property
    def version(self):              return self._version
    @property
    def matrix(self):               return self._matrix
    @property
    def gap_open(self):             return int(self._gap_open)
    @property
    def gap_extend(self):           return int(self._gap_extend)
    @property
    def database(self):             return self._db
    @property
    def lastiteration(self):        return int(self._lastiteration)
    @property
    def are_hits_corrected(self):    return self._correctedHits


    def get_hits(self, iteration = None, evalue = None, tz_type = None, tz_parameter = 0, overlap = 1):
        """
        Browse through the BlastHit of the BlastResult

        @type  iteration:    Integer
        @param iteration:    Iteration of interest. B{DEFAULT = last iteration}.
                             IF iteration is not a number returns as with DEFAULT.
                             IF iteration is out of range returns as with DEFAULT.
                             IF iteration == 0 returns the FIRST iteration

        @type  evalue:       Float
        @param evalue:       Maximum e-value threshold. B{DEFAULT = No e-value threshold}.

        @type  tz_type:      String
        @param tz_type:      Rost Evaluation Curve. B{DEFAULT = NO-ACTIVE}. I{OPTIONS = "ID" for identity curve
                                                                                        "SIM" for similarity curve}.

        @type  tz_parameter: Integer
        @param tz_parameter: Rost Evaluation Curve "Strength" Parameter. B{DEFAULT = 0}.

        @type  overlap:      Float
        @param overlap:      Amount of allowed overlap (0<->1) between single results. B{DEFAULT = 1}.

        @rtype: Array of L{BlastHist}

        B{TIP:} Rost Evaluation Curve is based on Rost's twilight zone:
                I{Rost, B. (1999). Twilight zone of protein sequence alignments. Protein Engineering, 12(2), 85 94.}
        """

        hits_of_interest = []

        '''
            First we check that the iteration asked is correct.
            Any error will redirect to either the first or last iteration (depending on the error)
        '''
        try: iteration = int(iteration)
        except: iteration = self.lastiteration

        if iteration == None:
            iteration = self.lastiteration
        elif iteration > self.lastiteration:
            iteration = self.lastiteration
        elif iteration == 0:
            iteration = 1

        '''
            We check that the asked options for the Rost curve are OK
            When the type is set to None, the curve is not applyed, even if tz_parameter is defined
        '''
        if tz_type != None:
            if not (tz_type == "ID" or tz_type == "SIM"):
                tz_type = None

        for hit in self._hits:
            '''
                We check out only hits of the iteration of interest
            '''
            if hit.iteration == iteration:
                '''
                    If applyes, we check the evalue
                '''
                if (evalue == None or hit.e_value <= float(evalue)):
                    '''
                        If applyes, we check Rost's curves
                    '''
                    if(tz_type == None or hit.evaluate_Rost_twilight_zone(equation = tz_type, parameter = tz_parameter)):
                        hits_of_interest.append(hit)
            '''
                As the hits are ordered, once we are over the iteration of interest, we can skip
            '''
            if hit.iteration > iteration:
                break

        if overlap == 1 or len(hits_of_interest) == 0:
            return hits_of_interest
        else:
            return self._get_nonoverlaping_hits(hitlist = hits_of_interest, overlap = overlap)

    def get_used_hits(self):
        """
        @rtype: Array of L{BlastHist}
        """
        hits_of_interest = []
        for hit in self._hits:
            if hit.is_used:
                hits_of_interest.append(hit)
        return hits_of_interest

    def _get_nonoverlaping_hits(self, hitlist, overlap = 1):
        """
        Due to the way this needs to be calculated, it will be a last step of the get_hits() function
        Returns all hits whose overlap with the already accepted ones is LOWER OR EQUAL than the given parameter
        It asumes that the hits are ordered by reliability, thus gives priority to the initial ones

        @type  hitlist: Array of L{BlastHist}
        @param hitlist: List of BlastHit to filter

        @type  overlap: Float
        @param overlap: Amount of allowed overlap (0<->1) between single results. B{DEFAULT = 1}.

        @rtype: Array of L{BlastHist}

        """
        final_hits = []
        final_hits.append(hitlist[0])
        for x in range(1,len(hitlist)):
            accepted = True
            for y in range(0,len(final_hits)):
                if hitlist[x].overlap(blastHit = final_hits[y]) > overlap:
                    accepted = False
                    break
            if accepted:
                final_hits.append(hitlist[x])

        return final_hits

    '''
        FUNCTIONS
    '''
    def correct_hit_count(self, countFile = None, countqueryFile = None, return_correction_dict = False):
        """
        Corrects the starting point of the hits.

        B{RATIONALE:}
        When blasting vs. PDB (for example), sometimes the hit positions given by blast are wrong,
        as the blast always consider the first position of the hit sequence as 1 and PDB does not.
        Given a tabulated file with:
        C{HIT_ID    FIRST_POSITION_REAL_NUMBER}
        (obviously the HIT_ID should match the format of the blasted database)

        @type  countFile:              String
        @param countFile:              Name of the "correction file"

        @type  return_correction_dict: Boolean
        @param return_correction_dict: When active the correction is not applied over the BlastHit objects and
                                       correctedHits is not set to True. The correction dictionary is returned

        @rtype:  Dictionary {String}[Integer]
        @return: Only when it is asked for through the return_correction_dict parameters

        @raise IOError:        IF the correction index file does not exist.
        @raise AttributeError: IF the BlastResult does not contain any BlastHit.
        @raise BlastError:     IF L{correctHitCount} has been called already.
        """

        if not os.path.isfile(countFile):
            raise IOError("The file %s does not exist and can not be used to correct the BlastHit positions.\n" %countFile)

        if not self.has_hits():
            warnings.warn("BlastResult of {0} has no hits to correct\n".format(self._query))

        if self.are_hits_corrected:
            raise BE(code = 3, value = "HitCounts has already been corrected\n")

        codes_of_interest = set([hit.sequenceID for hit in self._hits])
        if countqueryFile == countFile:
            codes_of_interest.add(self._query)
        start_index_dic = {}
        file_fd = open(countFile,'r')
        for line in file_fd:
            if len(line.strip()) > 0:
                if line.split('\t')[0].lstrip('>') in codes_of_interest:
                    start_index_dic[line.split('\t')[0].lstrip('>')] = line.split('\t')[1].strip().split(';')
        file_fd.close()

        if countqueryFile is not None and countqueryFile != countFile:
            file_fd = open(countqueryFile,'r')
            for line in file_fd:
                if len(line.strip()) > 0:
                    if line.split('\t')[0] == self._query:
                        start_index_dic[line.split('\t')[0]] = line.split('\t')[1].strip().split(';')
            file_fd.close()

        if return_correction_dict: return start_index_dic


        for hit in self._hits:
            '''
            This tests between the options PDB/PDB_ID or PDB_ID in case the TAB file has different codification
            '''
            try:
                hit.correct_hit_count(new_index = start_index_dic[hit.sequenceID])
                if countqueryFile is not None:
                    hit.correct_query_count(new_index = start_index_dic[self._query])
            except:
                hit_ID = hit.sequenceID.split("/")[-1]
                hit.correct_hit_count(new_index = start_index_dic[hit_ID])
                if countqueryFile is not None:
                    hit.correct_query_count(new_index = start_index_dic[self._query])
        self._correctedHits = True

    def set_last_iteration(self):
        """
        Modifies the value of self.lastiteration with the value of iteration of the last BlastHit found

        IF self.hits == [] : NO BLAST HITS HAVE BEEN FOUND then self.lastiteration remains at 0
        """
        if self._hits == []:
            self._lastiteration = 0
        else:
            self._lastiteration = self._hits[-1].iteration

    def has_hits(self):
        """
        @rtype: Boolean
        """
        if self.lastiteration == 0: return False
        else:                       return True

    def is_empty(self):
        """
        @rtype: Boolean
        """
        return not self.has_hits()

    '''
        BlastResult comparisson
    '''
    def same_blast_version(self, blastObject = None):
        """
        @type  blastObject: L{BlastResult}
        @param blastObject: BlastResult to which compare

        @rtype: Boolean

        @raise AttributeError: IF no blastObject is given
        """
        if blastObject == None:
            raise AttributeError("A second object needs to be given for the comparison\n")

        if self.version == blastObject.version: return True
        return False

    def same_blast_database(self, blastObject = None):
        """
        @type  blastObject: L{BlastResult}
        @param blastObject: BlastResult to which compare

        @rtype: Boolean

        @raise AttributeError: IF no blastObject is given
        """
        if blastObject == None:
            raise AttributeError("A second object needs to be given for the comparison\n")

        if self.database == blastObject.database: return True
        return False

    def same_blast_matrix(self, blastObject = None):
        """
        @type  blastObject: L{BlastResult}
        @param blastObject: BlastResult to which compare

        @rtype: Boolean

        @raise AttributeError: IF no blastObject is given
        """
        if blastObject == None:
            raise AttributeError("A second object needs to be given for the comparison\n")

        if self.matrix == blastObject.matrix: return True
        return False

    def same_blast_iterations(self, blastObject = None):
        """
        @type  blastObject: L{BlastResult}
        @param blastObject: BlastResult to which compare

        @rtype: Boolean

        @raise AttributeError: IF no blastObject is given
        """
        if blastObject == None:
            raise AttributeError("A second object needs to be given for the comparison\n")

        if self.lastiteration == blastObject.lastiteration: return True
        return False

    def same_blast_gap_properties(self, blastObject = None):
        """
        @type  blastObject: L{BlastResult}
        @param blastObject: BlastResult to which compare

        @rtype: Boolean

        @raise AttributeError: IF no blastObject is given
        """
        if blastObject == None:
            raise AttributeError("A second object needs to be given for the comparison\n")

        if (self.gap_open == blastObject.gap_open) and (self.gap_extend == blastObject.gap_extend): return True
        return False

    def comparable(self, blastObject = None):
        """
        @type  blastObject: L{BlastResult}
        @param blastObject: BlastResult to which compare

        @rtype: Boolean

        @raise AttributeError: IF no blastObject is given
        """
        if blastObject == None:
            raise AttributeError("A second object needs to be given for the comparison\n")

        if self.same_blast_database(blastObject   = blastObject)  and self.same_blast_version(blastObject        = blastObject) and \
           self.same_blast_matrix(blastObject     = blastObject)  and self.same_blast_gap_properties(blastObject = blastObject) and \
           self.same_blast_iterations(blastObject = blastObject): return True
        return False

    '''
        TO_STRING
    '''
    def str_compacted_blast(self, iteration = None, evalue = None, tz_type = None, tz_parameter = 0, overlap = 1):
        """
        @type  iteration:    Integer
        @param iteration:    Iteration of interest. B{DEFAULT = last iteration}.
                             IF iteration is not a number returns as with DEFAULT.
                             IF iteration is out of range returns as with DEFAULT.
                             IF iteration == 0 returns the FIRST iteration

        @type  evalue:       Float
        @param evalue:       Maximum e-value threshold. B{DEFAULT = No e-value threshold}.

        @type  tz_type:      String
        @param tz_type:      Rost Evaluation Curve. B{DEFAULT = NO-ACTIVE}. I{OPTIONS = "ID" for identity curve
                                                                                        "SIM" for similarity curve}.

        @type  tz_parameter: Integer
        @param tz_parameter: Rost Evaluation Curve "Strength" Parameter. B{DEFAULT = 0}.

        @type  overlap:      Float
        @param overlap:      Amount of allowed overlap (0<->1) between single results. B{DEFAULT = 1}.

        @rtype: String

        B{TIP:} Rost Evaluation Curve is based on Rost's twilight zone:
                I{Rost, B. (1999). Twilight zone of protein sequence alignments. Protein Engineering, 12(2), 85 94.}
        """

        lines = []

        for hit in self.get_hits(iteration=iteration, evalue=evalue, tz_type=tz_type, tz_parameter=tz_parameter, overlap=overlap):
            lines.append("\t".join((self.query,str(self.query_length),str(hit))))

        return "\n".join(lines)

    def str_PIR(self, iteration = None, result = 0, evalue = None, tz_type = None, tz_parameter = 0, overlap = 1):
        """
        @type  iteration:    Integer
        @param iteration:    Iteration of interest. B{DEFAULT = last iteration}.
                             IF iteration is not a number returns as with DEFAULT.
                             IF iteration is out of range returns as with DEFAULT.
                             IF iteration == 0 returns the FIRST iteration

        @type  result:       Integer
        @param result:       Result of interest (in order of selected results)

        @type  evalue:       Float
        @param evalue:       Maximum e-value threshold. B{DEFAULT = No e-value threshold}.

        @type  tz_type:      String
        @param tz_type:      Rost Evaluation Curve. B{DEFAULT = NO-ACTIVE}. I{OPTIONS = "ID" for identity curve
                                                                                        "SIM" for similarity curve}.

        @type  tz_parameter: Integer
        @param tz_parameter: Rost Evaluation Curve "Strength" Parameter. B{DEFAULT = 0}.

        @type  overlap:      Float
        @param overlap:      Amount of allowed overlap (0<->1) between single results. B{DEFAULT = 1}.

        @rtype: String

        B{TIP:} Rost Evaluation Curve is based on Rost's twilight zone:
                I{Rost, B. (1999). Twilight zone of protein sequence alignments. Protein Engineering, 12(2), 85 94.}
        """
        lines = []
        hit = self.get_hits(iteration=iteration, evalue=evalue, tz_type=tz_type, tz_parameter=tz_parameter, overlap=overlap)[result]

        #Counting in the possibility that the ID of the PDB hit is:
        #    PDB/2ERD_A
        hit_name = hit.sequenceID.replace("/","|").upper()
        hit_chain = hit_name.split("_")[-1]
        hit_id = hit_name.split("|")[-1].split("_")[0]
        hit_sq = hit.hit_seq + "*"
        q_sq = hit.query_seq + "*"

        header1 = ">P1;%s\nsequence:%s:%d:.:%d:.:.:.:.:." % (self.query,self.query,int(hit.query_pos[0]),int(hit.query_pos[-1]))
        header2 = ">P1;%s\nstructureX:%s:%d:%s:%d:%s:.:.:.:." % (hit_id,hit_id,int(hit.hit_pos[0]),hit_chain,int(hit.hit_pos[-1]),
                                                                 hit_chain)

        lines.append(header1)
        lines.extend([q_sq[i:i+60] for i in range(0, len(q_sq), 60)])
        lines.append(header2)
        lines.extend([hit_sq[i:i+60] for i in range(0, len(hit_sq), 60)])

        return "\n".join(lines)

    def print_query_fasta(self):
        """
        @rtype: String. B{Through STDOUT}.

        @raise IOError: IF query_sequence is not known.
        """

        if self.query_sequence == "":
            raise IOError("The sequence for %s is not known and cannot be printed\n" %self.query)

        print(">%s\n%s" %(self.query, self.query_sequence))

    def print_compacted_blast(self, iteration = None, evalue = None, tz_type = None, tz_parameter = 0, overlap = 1, file = None):
        """
        @type  iteration:    Integer
        @param iteration:    Iteration of interest. B{DEFAULT = last iteration}.
                             IF iteration is not a number returns as with DEFAULT.
                             IF iteration is out of range returns as with DEFAULT.
                             IF iteration == 0 returns the FIRST iteration

        @type  evalue:       Float
        @param evalue:       Maximum e-value threshold. B{DEFAULT = No e-value threshold}.

        @type  tz_type:      String
        @param tz_type:      Rost Evaluation Curve. B{DEFAULT = NO-ACTIVE}. I{OPTIONS = "ID" for identity curve
                                                                                        "SIM" for similarity curve}.

        @type  tz_parameter: Integer
        @param tz_parameter: Rost Evaluation Curve "Strength" Parameter. B{DEFAULT = 0}.

        @type  overlap:      Float
        @param overlap:      Amount of allowed overlap (0<->1) between single results. B{DEFAULT = 1}.

        @type  file:         String
        @param file:         Name of the output file. Redirects function output to file

        @rtype: String. B{Through STDOUT} IF file is None.
                String. B{Through FILE} otherwise.

        B{TIP:} Rost Evaluation Curve is based on Rost's twilight zone:
                I{Rost, B. (1999). Twilight zone of protein sequence alignments. Protein Engineering, 12(2), 85 94.}
        """
        if file:
            try:
                file_out = open(file, 'w')
            except:
                raise IOError("%s cannot be opened to write\n" %file)
            file_out.write("%s\n" % self.str_compacted_blast(iteration=iteration, evalue=evalue, tz_type=tz_type, tz_parameter=tz_parameter, overlap=overlap))
            file_out.close()
        else:
            print(self.str_compacted_blast(iteration=iteration, evalue=evalue, tz_type=tz_type, tz_parameter=tz_parameter, overlap=overlap))

    def str_blast_details(self):
        """
        @rtype: String
        """
        lines = []

        lines.append("Blast Version: %s" %self.version)
        lines.append("Search on matrix: %s" %self.matrix)
        lines.append("Gap open penalty: %s\nGap extension penalty: %s" %(str(self.gap_open),str(self.gap_extend)))
        lines.append("Database searched: %s\n" %self.database)

        return "\n".join(lines)

    def __str__(self):
        """
        Strings the String version of ALL the BlastHit objects contained (with all iterations!!)

        @rtype: String
        """
        lines = []

        for hit in self._hits:
            lines.append("\t".join((self.query,str(self.query_length),str(hit))))

        return "\n".join(lines)



class CDhitHomolog(object):
    def __init__(self, name, length, homology):
        self._name     = name.lstrip('>').rstrip('.')
        self._length   = int(length.replace('aa','').rstrip(','))
        self._homology = homology.replace('%','')

    @property
    def name(self):        return self._name
    @property
    def length(self):      return self._length
    @property
    def homology(self): 
        if self.is_master: return self._homology
        else:              return int(self._homology)
    @property
    def is_master(self):   return self._homology == '*'

    def __repr__(self):
        if not self.is_master:
            return '{0.name}: {0.length:0004d} Aa with {0.homology:003d}%'.format(self)
        else:
            return '{0.name}: {0.length:0004d} Aa.'.format(self)



from SBILib              import SBIglobals
from .                import process_DBREF_line
from SBILib.beans.JSONer import JSONer


class DBref(JSONer):

    '''
    {DBRef} includes the reference data to the PDB-CHAIN from:
    Known database-references include:

    Database name                            database (code in columns 27 - 32)
    ---------------------------------------------------------------------------
    BioMagResBank                            BMRB
    BLOCKS                                   BLOCKS
    European Molecular Biology Laboratory    EMBL
    GenBank                                  GB
    Genome Data Base                         GDB
    Nucleic Acid Database                    NDB
    PROSITE                                  PROSIT
    Protein Data Bank                        PDB
    Protein Identification Resource          PIR
    SWISS-PROT                               SWS
    TREMBL                                   TREMBL
    UNIPROT                                  UNP

    '''
    valid_references = {
        'BMRB'      : 'BioMagResBank',
        'BLOCKS'    : 'BLOCKS',
        'EMBL'      : 'European Molecular Biology Laboratory',
        'GB'        : 'GenBank',
        'GDB'       : 'Genome Data Base',
        'NDB'       : 'Nucleic Acid Database',
        'PROSIT'    : 'PROSITE',
        'PDB'       : 'Protein Data Bank',
        'PIR'       : 'Protein Identification Resource',
        'SWS'       : 'SWISS-PROT',
        'TREMBL'    : 'TREMBL',
        'UNP'       : 'UNIPROT'
    }

    def __init__(self, pdb_line):
        '''
        @type  pdb_line:     String
        @param pdb_pdb_line: Line of a PDB. Starts with DBREF

        @raise AttributeError if line does not start with DBREF
        '''
        if not pdb_line.startswith('DBREF '):
            SBIglobals.error(self, '{0} cannot create DBref'.format(pdb_line))
        data          = self._process_line(pdb_line)

        self._pdb     = data[0]
        self._chain   = data[1]
        self._start   = data[2]
        self._end     = data[3]
        self._db      = data[4]
        self._ref     = data[5]

    #
    # ATTRIBUTES
    #
    @property
    def pdb(self):
        return self._pdb

    @property
    def chain(self):
        return self._chain

    @property
    def start(self):
        return int(self._start)

    @property
    def idxs(self):
        return self._start.index

    @property
    def end(self):
        return int(self._end)

    @property
    def idxe(self):
        return self._end.index

    @property
    def database(self):
        return self.valid_references[self._db]

    @property
    def reference(self):
        return self._ref

    #
    # BOOLEANS
    #
    @property
    def is_reference_to_db(self, db_minicode):
        db_minicode = db_minicode.upper()
        if not db_minicode in self.valid_references:
            line1 = '{0} is not a valid DBref code'.format(db_minicode)
            line2 = 'Available codes: {0}'.format(list(self.valid_references.keys()))
            SBIglobals.error(self, "\n".join([line1, line2]))

        #SWS and TREMBL are parts of UNP
        if db_minicode == 'UNP' and self._db in ['SWS', 'TREMBL']:
            return True
        return self._db == db_minicode

    #
    # PRIVATE
    #
    def _process_line(self, pdb_line):
        return process_DBREF_line(pdb_line)

    #
    # OUT FUNCTIONS
    #
    def as_dict(self):
        nobj = {}
        nobj['chain']     = self.chain
        nobj['db']        = self.database
        nobj['start']     = self.start
        nobj['idxs']      = self.idxs
        nobj['end']       = self.end
        nobj['idxe']      = self.idxe
        nobj['reference'] = self.reference
        return nobj

    def __repr__(self):
        return repr(self.json())



"""
ContactAA

author: jbonet
date:   02/2014

@oliva's lab
"""

from .   import Contact
from SBILib import SBIglobals

class ContactAA(Contact):

    available_distance_types = set(["min","ca","cb","geometric","backbone"])
    atomic_types             = set(["min","ca","cb","backbone"])
    description              = 'AminoAmino'

    def __init__(self, aminoacid1, aminoacid2, threshold_type = "cb", threshold_distance = 12):
        
        self._check_type(requested_type = threshold_type)

        self._threshold_type     = threshold_type
        self._threshold_distance = threshold_distance

        super(ContactAA, self).__init__(residue1 = aminoacid1, residue2 = aminoacid2)

    #
    # ATTRIBUTES
    #
    @property
    def aminoacid1(self):         return self._residue1

    @property
    def aminoacid2(self):         return self._residue2

    @property
    def min_distance(self):       return self._distance["min"][2]
    @property
    def min_atoms(self):          return self._distance["min"][0:2]

    @property
    def ca_distance(self):        return self._distance["ca"][2]
    @property
    def ca_atoms(self):           return self._distance["ca"][0:2]

    @property
    def cb_distance(self):        return self._distance["cb"][2]
    @property
    def cb_atoms(self):           return self._distance["cb"][0:2]

    @property
    def geometric_distance(self): return self._distance["geometric"][2]

    @property
    def backbone_distance(self):  return self._distance["backbone"][2]
    @property
    def backbone_atoms(self):     return self._distance["backbone"][0:2]

    #
    # OVERWRITE PARENT FUNCTION
    #
    def _build(self):
        SBIglobals.alert('deepdebug', self, 'Analyzing AA Contact {0.type}:{0.number} - {1.type}:{1.number}'.format(self.aminoacid1, self.aminoacid2))

        for dist_type in self.available_distance_types:
            self._distance.setdefault(dist_type, None)

        self._distance[self._threshold_type] = self.aminoacid1.distance(self.aminoacid2, dist_type = self._threshold_type)
        SBIglobals.alert('deepdebug', self, '\tEvaluating distance {0:.3f} of {1}'.format(self._distance[self._threshold_type][2], self._threshold_type))
        if float(self._distance[self._threshold_type][2]) <= self._threshold_distance and \
           float(self._distance[self._threshold_type][2]) >= 0:
            SBIglobals.alert('deepdebug', self, '\tDistance under threshold.')
            self._underthreshold = True
            for dist_type in self._distance:
                if dist_type != self._threshold_type:
                    SBIglobals.alert('deepdebug', self, '\tGathering {0} distance'.format(dist_type))
                    self._distance[dist_type] = self.aminoacid1.distance(self.aminoacid2, dist_type = dist_type)
                    SBIglobals.alert('deepdebug', self, '\t\tDistance {0:.3f} of {1}'.format(self._distance[dist_type][2], dist_type))      

    #
    # TOSTRING
    #
    def toString(self, requested_type = "cb"):
        if requested_type != "all":
            self._check_type(requested_type = requested_type)
            ddata = self._distance[requested_type]
            if requested_type in self.atomic_types:
                return '{0.aminoacid1.type}:{0.aminoacid1.number}:{1[0].name}\t{0.aminoacid2.type}:{0.aminoacid2.number}:{1[1].name}\t{1[2]:.3f}'.format(self, ddata)
            else:
                return '{0.aminoacid1.type}:{0.aminoacid1.number}\t{0.aminoacid2.type}:{0.aminoacid2.number}\t{1[2]:.3f}'.format(self, ddata)
        else:
            return '{0.aminoacid1.type}:{0.aminoacid1.number}\t{0.aminoacid2.type}:{0.aminoacid2.number}\t{0.min_distance:.3f}\t{0.ca_distance:.3f}\t{0.cb_distance:.3f}\t{0.geometric_distance:.3f}\t{0.backbone_distance:.3f}'.format(self)
        



"""
PPInterface

author: jbonet
date:   03/2013

@oliva's lab
"""

from .         import Interface
from ..contact import ContactAA

class PPInterface(Interface):
    """
    {Interface} between two protein chains
    """
    def __init__(self, protein_chain   = None, protein_interactor = None, 
                       threshold_type  = "cb", threshold_distance = 12,
                       protein_centers = None, interactor_centers = None):

        if protein_centers is None:
            protein_centers    = protein_chain.geometric_center(structure = True,  hetero     = False,
                                                                water     = False, by_residue = True)
        if interactor_centers is None:
            interactor_centers = protein_interactor.geometric_center(structure = True,  hetero     = False,
                                                                     water     = False, by_residue = True)

        super(PPInterface, self).__init__(chain1         = protein_chain,   chain2             = protein_interactor,
                                          threshold_type = threshold_type,  threshold_distance = threshold_distance,
                                          chain1_centres = protein_centers, chain2_centers     = interactor_centers)
        
    #
    # ATTRIBUTES
    #
    @property
    def protein_chain(self):      return self._chain1
    @property
    def protein_id(self):         return self._chain1.globalID
    @property
    def protein_centers(self):    return self._chain1_centers

    @property
    def protein_interactor(self): return self._chain2
    @property
    def interactor_id(self):      return self._chain2.globalID
    @property
    def interactor_centers(self): return self._chain1_centers

    @property
    def protein_positions(self):
        return self._list_positions(1)
    @property 
    def protein_view_interface(self):
        return self._view_interface_from(1)

    @property
    def interactor_positions(self):
        return self._list_positions(2)
    @property 
    def interactor_view_interface(self):
        return self._view_interface_from(2)

    #
    # OVERWRITE PARENT: PRIVATE FUNCTIONS
    #
    def _build(self):
        if len(self.protein_chain.aminoacids) == 0 or len(self.protein_interactor.aminoacids) == 0: return

        if (self.protein_chain.is_only_ca or self.protein_interactor.is_only_ca) and self.threshold_type == "cb":
            self._threshold_type     = "ca"
            # Add distance of two C-C bonds (1.54A) -> http://en.wikipedia.org/wiki/Bond_length
            # Gives margin by ajusting at 2
            self._threshold_distance = self.threshold_distance + (2*2)

        super(PPInterface, self)._build()

        for i in range(len(self._filtered[0])):
            new_contact = ContactAA(aminoacid1         = self.protein_chain.aminoacids[self._filtered[0][i]], 
                                    aminoacid2         = self.protein_interactor.aminoacids[self._filtered[1][i]], 
                                    threshold_type     = self.threshold_type, 
                                    threshold_distance = self.threshold_distance)
            if new_contact.is_underthreshold:
                    self.contacts = new_contact

    #
    # OVERWRITE PARENT: TOSTRING
    #
    def toString(self, all_types = False):
        data = []
        if not all_types:
            data.append('{0._chain1.chain}\t{0._chain2.chain}\t{0.threshold_type}\t{0.threshold_distance}'.format(self))
        else:
            data.append('{0._chain1.chain}\t{0._chain2.chain}\tmin\tca\tcb\tgeometric\tbackbone'.format(self))
        data.append(super(PPInterface, self).toString(all_types))
        return "\n".join(data)


        # for r1 in self.chain.aminoacids + self.chain.heteroatoms:
        #     for r2 in self.interactor.aminoacids + self.chain.heteroatoms:
        #         new_contact = r1.distance(residue = r2, dist_type = self.c_type)
        #         if new_contact.distance <= self.max_distance:
        #             if self._alldistances:
        #                 if   self.c_type == "cb":  altd = ["ca","min"]
        #                 elif self.c_type == "ca":  altd = ["cb","min"]
        #                 elif self.c_type == "min": altd = ["cb","ca"]
        #                 new_contact._distance[1] = r1.distance(residue = r2, dist_type = altd[0]).distance
        #                 new_contact._distance[2] = r1.distance(residue = r2, dist_type = altd[1]).distance
        #             self.add_contact(contact  = new_contact)

    # def __str__(self):
    #     if not self._alldistances:
    #         return super(PPInterface, self).__str__()
    #     else :
    #         data = []
    #         # data.append('CHAIN: {0.chain_id}'.format(self))
    #         for contact in self.contacts:
    #             data.append(contact.print_aldist())
    #         return "\n".join(data)

# from ContactList import ContactList

# class PPInterface(ContactList):
#     """
#     {ContactList} between two chains
#     """
#     def __init__(self, chain = None, interactor = None, c_type = "cb", max_distance = 12, all_distances = False):
#         self._alldistances = all_distances
#         super(PPInterface, self).__init__(chain = chain, secondary_chain = interactor,
#                                           c_type = c_type, max_distance = max_distance)
        

#     #
#     # ATTRIBUTES
#     #
#     @property
#     def interactor(self):
#         """
#         Structure of the interactor chain
#         @rtype: {CHAIN}
#         """
#         return self._schain

#     @property
#     def interactor_id(self):
#         """
#         Identifier of interactor chain
#         @rtype: String
#         """
#         return self._schain.globalID()

#     @property
#     def chain_positions(self):
#         """
#         Positions in chain1 related to the interaction
#         [OVERWRITTED FROM PARENT]
#         @rtype: set
#         """
#         positions = set()
#         for contact in self._contacts:
#             positions.add(contact.residue1.number)
#         return positions

#     @property
#     def interactor_positions(self):
#         """
#         Positions in chain1 related to the interaction
#         @rtype: set
#         """
#         positions = set()
#         for contact in self._contacts:
#             positions.add(contact.residue2.number)
#         return positions

#     #
#     # BOOLEANS
#     #
#     def fulfills_interface(self, interface, strict = False):
#         """
#         Checks if a list of contacts appear in the interface

#         @type  interface: {Interface}
#         @param contact_list: List of {Contact}s to check

#         @type    strict: Boolean
#         @param   strict: Strict checking... all the given contacts must exist
#         @default strict: False

#         @rtype: Boolean
#         """
#         total = 0

#         for c1 in self.contacts:
#             for c2 in interface.contacts:
#                 if c1 == c2:
#                     total += 1

#         if total == len(interface):  return True
#         if not strict and total > 0: return True
#         return False


#     def fulfills_exclude_residues_chain(self, residue_number_list, strict = False):
#         """
#         Checks if a list of residues (number) DOES NOT appear in the chain

#         @type  residue_number_list: List of Integers
#         @param residue_number_list: List of residue numbers to check

#         @type    strict: Boolean
#         @param   strict: Strict checking... EVERY REQUESTED RESIDUE HAS NOT TO BE IN THE INTERFACE
#         @default strict: False

#         @rtype: Boolean
#         """
#         return not self._fulfills_residues(residue_number_list = residue_number_list, chain = True, strict = strict)

#     def fulfills_residues_chain(self, residue_number_list, strict = False):
#         """
#         Checks if a list of residues (number) appear in the chain

#         @type  residue_number_list: List of Integers
#         @param residue_number_list: List of residue numbers to check

#         @type    strict: Boolean
#         @param   strict: Strict checking... EVERY REQUESTED RESIDUE HAS TO BE IN THE INTERFACE
#         @default strict: False

#         @rtype: Boolean
#         """
#         return self._fulfills_residues(residue_number_list = residue_number_list, chain = True, strict = strict)

#     def fulfills_exclude_residues_interactor(self, residue_number_list, strict = False):
#         """
#         Checks if a list of residues (number) DOES NOT appear in the interactor

#         @type  residue_number_list: List of Integers
#         @param residue_number_list: List of residue numbers to check

#         @type    strict: Boolean
#         @param   strict: Strict checking... EVERY REQUESTED RESIDUE HAS NOT TO BE IN THE INTERFACE
#         @default strict: False

#         @rtype: Boolean
#         """
#         return not self._fulfills_residues(residue_number_list = residue_number_list, interactor = True, strict = strict)

#     def fulfills_residues_interactor(self, residue_number_list, strict = False):
#         """
#         Checks if a list of residues (number) appear in the interactor

#         @type  residue_number_list: List of Integers
#         @param residue_number_list: List of residue numbers to check

#         @type    strict: Boolean
#         @param   strict: Strict checking... EVERY REQUESTED RESIDUE HAS TO BE IN THE INTERFACE
#         @default strict: False

#         @rtype: Boolean
#         """
#         return self._fulfills_residues(residue_number_list = residue_number_list, interactor = True, strict = strict)

#     def _fulfills_residues(self, residue_number_list, chain = False, interactor = False, strict = False):
#         """
#         Checks if a list of residues (number) appear in a given chain

#         @type  residue_number_list: List of Integers
#         @param residue_number_list: List of residue numbers to check

#         @type    chain: Boolean
#         @param   chain: Check on chain?
#         @default chain: False

#         @type    interactor: Boolean
#         @param   interactor: Check on interactor?
#         @default interactor: False

#         @type    strict: Boolean
#         @param   strict: Strict checking... EVERY REQUESTED RESIDUE HAS TO BE IN THE INTERFACE
#         @default strict: False

#         @rtype: Boolean

#         @raise: AttributeError if chain and interactor are both False
#         """
#         if residue_number_list is None: return True

#         if not chain and not interactor:
#             raise AttributeError()

#         if chain:
#             set_of_interest = self.chain_positions
#         if interactor:
#             set_of_interest = self.interactor_positions

#         set_of_requested = set([int(x) for x in residue_number_list])

#         intersection = set_of_interest.intersection(set_of_requested)

#         if len(intersection) == len(set_of_requested): return True
#         if not strict and len(intersection) > 0:       return True
#         return False

#     #
#     # OVERWRITE PARENT
#     #
#     def _build(self):
#         """
#         Builds the contacts between the two chains
#         """
#         for r1 in self.chain.aminoacids + self.chain.heteroatoms:
#             for r2 in self.interactor.aminoacids + self.chain.heteroatoms:
#                 new_contact = r1.distance(residue = r2, dist_type = self.c_type)
#                 if new_contact.distance <= self.max_distance:
#                     if self._alldistances:
#                         if   self.c_type == "cb":  altd = ["ca","min"]
#                         elif self.c_type == "ca":  altd = ["cb","min"]
#                         elif self.c_type == "min": altd = ["cb","ca"]
#                         new_contact._distance[1] = r1.distance(residue = r2, dist_type = altd[0]).distance
#                         new_contact._distance[2] = r1.distance(residue = r2, dist_type = altd[1]).distance
#                     self.add_contact(contact  = new_contact)

#     def __str__(self):
#         if not self._alldistances:
#             return super(PPInterface, self).__str__()
#         else :
#             data = []
#             # data.append('CHAIN: {0.chain_id}'.format(self))
#             for contact in self.contacts:
#                 data.append(contact.print_aldist())
#             return "\n".join(data)



# -*-
#
# @author: jaumebonet
# @email:  jaume.bonet@gmail.com
# @url:    jaumebonet.github.io
#
# @date:   2016-02-15 11:21:10
#
# @Last modified by:   bonet
# @Last modified time: 12-Oct-2018
#
# -*-
from .Element import PeriodicTable
e = PeriodicTable()


class WrongAtomCodingName(Exception):
    pass


class AtomVariants(object):
    """Translator between how atoms are called in different formats"""
    def __init__(self, names):
        self.bmrb  = names[0]
        self.pdb   = names[1]
        self.ucfs  = names[2]
        self.msi   = names[3]
        self.xplor = names[4]
        self.sybyl = names[5]
        self.midas = names[6]
        self.diana = names[7]

    @staticmethod
    def known_format(self, format):
        if not format.lower() in self.__dict__:
            raise WrongAtomCodingName

    def get_name(self, name):
        return self.__getattribute__(name.lower())

    def get_atom_type(self):
        return e[self.bmrb[0]]


C    = AtomVariants(['C',    'C',    '',     'C',    'C',    'C',    'C',   'C'])
CA   = AtomVariants(['CA',   'CA',   '',     'CA',   'CA',   'CA',   'CA',  'CA'])
CB   = AtomVariants(['CB',   'CB',   '',     'CB',   'CB',   'CB',   'CB',  'CB'])
CD   = AtomVariants(['CD',   'CD',   '',     'CD',   'CD',   'CD',   'CD',  'CD'])
CD1  = AtomVariants(['CD1',  'CD1',  '',     'CD1',  'CD1',  'CD1',  'CD1', 'CD1'])
CD2  = AtomVariants(['CD2',  'CD2',  '',     'CD2',  'CD2',  'CD2',  'CD2', 'CD2'])
CE   = AtomVariants(['CE',   'CE',   '',     'CE',   'CE',   'CE',   'CE',  'CE'])
CE1  = AtomVariants(['CE1',  'CE1',  '',     'CE1',  'CE1',  'CE1',  'CE1', 'CE1'])
CE2  = AtomVariants(['CE2',  'CE2',  '',     'CE2',  'CE2',  'CE2',  'CE2', 'CE2'])
CG   = AtomVariants(['CG',   'CG',   '',     'CG',   'CG',   'CG',   'CG',  'CG'])
CG2  = AtomVariants(['CG2',  'CG2',  '',     'CG2',  'CG2',  'CG2',  'CG2', 'CG2'])
CZ   = AtomVariants(['CZ',   'CZ',   '',     'CZ',   'CZ',   'CZ',   'CZ',  'CZ'])
H    = AtomVariants(['H',    'H',    'HN',   'HN',   'HN',   'H',    '',    'HN'])
HA   = AtomVariants(['HA',   'HA',   'HA',   'HA',   'HA',   'HA',   '',    'HA'])
HB   = AtomVariants(['HB',   'HB',   'HB',   'HB',   'HB',   'HB',   '',    'HB'])
HB1  = AtomVariants(['HB1',  '1HB',  'HB1',  'HB1',  'HB1',  'HB1',  '',    'HB1'])
HB2  = AtomVariants(['HB2',  '1HB',  'HB1',  'HB1',  'HB2',  'HB2',  '',    'HB2'])
HB2b = AtomVariants(['HB2',  '2HB',  'HB2',  'HB2',  'HB2',  'HB2',  '',    'HB2'])
HB3  = AtomVariants(['HB3',  '2HB',  'HB2',  'HB2',  'HB1',  'HB1',  '',    'HB3'])
HB3b = AtomVariants(['HB3',  '3HB',  'HB3',  'HB3',  'HB3',  'HB3',  '',    'HB3'])
HD1  = AtomVariants(['HD1',  'HD1',  'HD1',  'HD1',  'HD1',  'HD1',  '',    'HD1'])
HD11 = AtomVariants(['HD11', '1HD1', 'HD11', 'HD11', 'HD11', 'HD11', '',    'HD11'])
HD12 = AtomVariants(['HD12', '2HD1', 'HD12', 'HD12', 'HD12', 'HD12', '',    'HD12'])
HD13 = AtomVariants(['HD13', '3HD1', 'HD13', 'HD13', 'HD13', 'HD13', '',    'HD13'])
HD2  = AtomVariants(['HD2',  'HD2',  'HD2',  'HD2',  'HD2',  'HD2',  '',    'HD2'])
HD2b = AtomVariants(['HD2',  '1HD',  'HD1',  'HD1',  'HD2',  'HD2',  '',    'HD2'])
HD3  = AtomVariants(['HD3',  '2HD',  'HD2',  'HD2',  'HD1',  'HD1',  '',    'HD3'])
HE1  = AtomVariants(['HE1',  'HE1',  'HE1',  'HE1',  'HE1',  'HE1',  '',    'HE1'])
HE2  = AtomVariants(['HE2',  'HE2',  'HE2',  'HE2',  'HE2',  'HE2',  '',    'HE2'])
HE2b = AtomVariants(['HE2',  '',     '',     'HE2',   '',    '',     '',    'HE2']),
HG   = AtomVariants(['HG',   'HG',   'HSG',  'HG',   'HG',   'HG',   '',    'HG'])
HG2  = AtomVariants(['HG2',  '1HG',  'HG1',  'HG1',  'HG2',  'HG2',  '',    'HG2'])
HG3  = AtomVariants(['HG3',  '2HG',  'HG2',  'HG2',  'HG1',  'HG1',  '',    'HG3'])
HG21 = AtomVariants(['HG21', '1HG2', 'HG21', 'HG21', 'HG21', 'HG21', '',    'HG21'])
HG22 = AtomVariants(['HG22', '2HG2', 'HG22', 'HG22', 'HG22', 'HG22', '',    'HG22'])
HG23 = AtomVariants(['HG23', '3HG2', 'HG23', 'HG23', 'HG23', 'HG23', '',    'HG23'])
N    = AtomVariants(['N',    'N',    '',     'N',    'N',    'N',    'N',   'N'])
NE2  = AtomVariants(['NE2',  'NE2',  '',     'NE2',  'NE2',  'NE2',  'NE2', 'NE2'])
O    = AtomVariants(['O',    'O',    '',     'O',    'O',    'O',    'O',   'O'])
OD1  = AtomVariants(['OD1',  'OD1',  '',     'OD1',  'OD1',  'OD1',  'OD1', 'OD1'])
OE1  = AtomVariants(['OE1',  'OE1',  '',     'OE1',  'OE1',  'OE1',  'OE1', 'OE1'])
OE2  = AtomVariants(['OE2',  'OE2',  '',     'OE2',  'OE2',  'OE2',  'OE2', 'OE2'])
SG   = AtomVariants(['SG',   'SG',   '',     'SG',   'SG',   'SG',   'SG',  'SG'])



"""PDBeChemlink

author: jbonet
date:   10/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
import sys, os, re
import subprocess
import warnings
import urllib.request, urllib.parse, urllib.error

"""
Dependences in SBI library
"""
from SBILib               import SBIglobals
from SBILib.databases import PDBeChemftp
from SBILib.beans.Path     import Path
from SBILib.beans.File     import File
from SBILib.data      import element_dic

class PDBeChemlink(object):
    """The PDBeChemlink class controls the download and parsing of PDBeChem database

    """
    def __init__(self, local = None):
        self._local   = os.path.abspath(local)
        self.__name__ = 'databases.PDBeChemlink'    # This must be included in every class for the SBIglobals.alert()

    """ATTRIBUTES"""
    @property
    def local(self):         return self._local
    @local.setter
    def local(self, value):  self._local = os.path.abspath(value)

    @property
    def localPDBeChems(self):
        for chem_file in Path.list_files(root = self.local, pattern = '*.cif'):
            yield chem_file

    @property
    def source(self):
        return PDBeChemftp['show']

    """BOOLEANS"""
    @property
    def has_local(self):    return self._local is not None

    """METHODS"""
    def download(self):
        if not self.has_local:
            raise NameError('A local PDBeChem database directory must be defined.')

        Path.mkdir(self.local)
        destination = os.path.join(self.local, 'mmcif.tar.gz')
        try:
            urllib.request.urlretrieve(PDBeChemftp['global'], destination)
        except:
            return False
        command = ['tar', 'zxvf', destination, '-C', self.local]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        out, err = p.communicate()

        return True

    def get_PDBeChem(self, chemID):
        if self.has_local:
            for chem_file in self.localPDBeChems:
                newfile = File(file_name = chem_file, action = 'r')
                if newfile.prefix.upper() == chemID.upper():
                    return chem_file

        #If we do not find it in local (or we do not have a local) we search it on the FTP
        chem_file = chemID.upper() + '.cif'
        source = PDBeChemftp['single'] + chem_file
        try:
            urllib.request.urlretrieve(source, chem_file)
        except:
            return False
        return os.path.abspath(chem_file)

    def get_PDBeChems(self, chemIDset):
        if isintance(chemIDset, str):
            warnings.warn('For single PDBeChem search the get_PDBeChem function is recomended.')
            yield self.get_PDBeChem(chemIDset)
        else:
            chemIDset = set([x.upper() for x in chemIDset])

        if self.has_local:
            for chem_file in self.localPDBeChems:
                newfile = File(file_name = chem_file, action = 'r')
                if newfile.prefix.lstrip('pdb').upper() in chemIDset:
                    yield chem_file
        else:
            for chemID in chemIDset:
                yield self.get_PDBeChem(chemID)

class PDBeChem(object):
    """
    """
    def __init__(self, cif_file):
        self._file    = File(file_name = cif_file, action = 'r')
        self.__name__ = 'databases.PDBeChem'    # This must be included in every class for the SBIglobals.alert()

        self._id        = None
        self._name      = None
        self._type      = None
        self._formula   = None
        self._parent    = None
        self._weight    = None
        self._fcharge   = None
        self._code1l    = None
        self._flformula = {}

        self._parse()
        self._decompose_formula()

    """ATTRIBUTES"""
    @property
    def id(self):            return self._id
    @property
    def name(self):          return self._name
    @property
    def type(self):          return self._type
    @property
    def formula(self):       return self._formula
    @property
    def full_formula(self):  return self._flformula
    @property
    def parent(self):        return self._parent
    @property
    def weight(self):        return self._weight
    @property
    def formal_charge(self): return self._fcharge
    @property
    def code1(self):         return self._code1l
    @property
    def code3(self):         return self._id

    """PRIVATE METHODS"""
    def _parse(self):
        for line in self._file.descriptor:
            if line.startswith('_chem_comp.'):
                line  = line.replace('_chem_comp.','')
                value = line[35:].strip().strip('"')
                value = value.replace(' (NON-PREFERRED NAME)','')
                value = value if value != '?' else None
                if line.startswith('id'):                      self._id      = value
                if line.startswith('pdbx_type'):               self._type    = value
                if line.startswith('formula '):                self._formula = value
                if line.startswith('formula_weight'):          self._weight  = value
                if line.startswith('pdbx_formal_charge'):      self._fcharge = value
                if line.startswith('one_letter_code'):         self._code1l  = value
                if line.startswith('name'):                    self._name    = value.upper()
                if line.startswith('mon_nstd_parent_comp_id'):
                    self._parent  = set([x.strip() for x in value.split(',')]) if value is not None else None
            if line.startswith(';') and self._name == '':      self._name  += line.strip().lstrip(';').upper()
        self._file.close()

    def _decompose_formula(self):
        if self.formula is not None:
            data    = self.formula.split()
            atregex = re.compile('(\D+)(\d*)')
            for atom in data:
                m = atregex.search(atom)
                if m.group(1) in element_dic:
                    self._flformula[m.group(1)] = m.group(2) if m.group(2) != '' else 1


    """OVERWRITE INHERITED FUNCTIONS"""
    def __str__(self):
        if self.code1 is not None and self.parent is not None:
            return "[{0.id} - {0.code1} from {0.parent}: {0.weight} - {0.formula} - {0.formal_charge}] {0.name} - {0.type}".format(self)
        elif self.code1 is not None:
            return "[{0.id} - {0.code1}: {0.weight} - {0.formula} - {0.formal_charge}] {0.name} - {0.type}".format(self)
        elif self.parent is not None:
            return "[{0.id} from {0.parent}: {0.weight} - {0.formula} - {0.formal_charge}] {0.name} - {0.type}".format(self)
        else:
            return "[{0.id}: {0.weight} - {0.formula} - {0.formal_charge}] {0.name} - {0.type}".format(self)





# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>

.. affiliation::
    Structural BioInformatics Lab <sbi.upf.edu>
    Baldo Oliva <baldo.oliva@upf.edu>

.. class:: CATHLink
"""

# Standard Libraries
from tempfile import NamedTemporaryFile
import os

# External Libraries
import requests
import pandas as pd

# This Library
import SBI.core as core


__all__ = ['CATHLink']


class CATHLink( object ):
    """Controls data retrieval from the CATH_ database.

    .. note::
        This functionality depends on the :ref:`global configuration options <configuration>` ``db.cath``,
        that points to the top level url of the downloadable files of the database. The link will target
        the latest release of  the database.

    .. _CATH: http://www.cathdb.info/
    """
    def __init__( self ):
        self.__name__ = 'databases.CATHLink'

    def get_data( self ):
        """Retrieve data from the declared SCOP version.

        :return: :class:`~pandas.DataFrame`
        """
        url = '/'.join([core.get_option('db', 'cath'), 'cath-classification-data', 'cath-domain-boundaries.txt'])
        local_file = _retrieve_data(url)
        df = pd.read_fwf(local_file.name, comment='#', header=None, delimiter='~',
                         names=['chain_name', 'domain_count', 'frag_count', 'data'],
                         colspecs=[(0, 5), (6, 9), (9, 13), (13, None)])
        df[['domain_count']] = df['domain_count'].str.extract(r'(\d+)')
        df['domain_count'] = pd.to_numeric(df['domain_count'])
        df['data'] = df.apply(lambda row: process_CATH_data(row), axis=1)
        df = _split_dataframe_rows(df.drop(columns=['frag_count']), ['data']).rename(columns={'data': 'selectors'})
        df['chain_name'] = chain_rename(df['chain_name'])
        df['pdb'] = df['chain_name'].str[:4]
        df = df.rename(columns={'chain_name': 'domain_id'}).drop(columns=['domain_count'])
        df['selectors'] = df['selectors'].map(tuple)
        return df


def chain_rename(column):
    count = 1
    last = ''
    new_column = []
    for c in column:
        if c != last:
            last = c
            count = 1
        else:
            count += 1
        new_column.append(c.strip() + '{:02d}'.format(count))
    return new_column


def process_CATH_data( row ):
    total = row['domain_count']
    data = row['data'].split()
    domains = []
    segments = []
    status = 'count'
    new_status_count = 0
    positions = []
    for i, d in enumerate(data):
        if i == new_status_count:
            status = 'count'
            if len(domains) == total:
                break
        if status == 'count':
            domains.append([[], ] * int(d))
            segments.append((int(d), i))
            status = 0
            new_status_count = i + (6 * int(d)) + 1
            for x in range(int(d)):
                positions.append((len(domains) - 1, x))
            continue
    for x in reversed(segments):
        data.pop(x[1])
    segments = [x[0] for x in segments]
    for i, x in enumerate(range(0, len(data), 6)):
        if i >= sum(segments):
            break
        cdata = data[x:x + 6]
        if cdata[0] != cdata[3]:
            raise Exception('whaaat?')
        strdata = '{0}:{1}{2};{3}:{4}{5}'.format(*cdata).replace('-;', ';').rstrip('-')
        strdata = '{0[0]}:{0[1]}-{1[1]}'.format(*[x.split(':') for x in strdata.split(';')])
        domains[positions[i][0]][positions[i][1]] = strdata
    return domains


def _split_dataframe_rows( df, columns=None ):
    # modified from: https://gist.github.com/jlln/338b4b0b55bd6984f883#gistcomment-2321628
    def _split_list_to_rows_inner(row, row_accumulator, column_selectors):
        split_rows = {}
        max_split = 0
        for column_selector in column_selectors:
            split_row = row[column_selector]
            split_rows[column_selector] = split_row
            if len(split_row) > max_split:
                max_split = len(split_row)

        for i in range(max_split):
            new_row = row.to_dict()
            for column_selector in column_selectors:
                try:
                    new_row[column_selector] = split_rows[column_selector].pop(0)
                except IndexError:
                    new_row[column_selector] = ''
            row_accumulator.append(new_row)

    columns = list(df.columns) if columns is None else columns
    new_rows = []
    df.apply(_split_list_to_rows_inner, axis=1, args=(new_rows, columns))
    new_df = pd.DataFrame(new_rows, columns=df.columns)
    return new_df


def _retrieve_data( url ):
    """Download the url into a tempfile.
    """
    if core.get_option('system', 'verbose') == 2:
        print('Downloading from {} '.format(url))
    local_file = NamedTemporaryFile(delete=core.get_option('system', 'verbose') < 2)
    r = requests.get(url, stream=True)
    if r.status_code != requests.codes.ok:
        raise IOError('Remote path {} cannot be found'.format(url))
    for chunk in r.iter_content(chunk_size=1024):
        if chunk:
            local_file.write(chunk)
    if core.get_option('system', 'verbose') == 2:
        print('into temporary file {}'.format(local_file.name))
    local_file.flush()
    os.fsync(local_file.fileno())
    return local_file



'''
Needleman-Wunsch global alignment algorithm

Based upon:
http://www.codesofmylife.com/
as defined in:
http://en.wikipedia.org/wiki/Needleman-Wunsch_algorithm

And adapted to the SBI library
A variation has been added so that initial movement of the alignment can
be penalized differently
'''
import numpy as np
import re

from SimilarityMatrix import SimilarityMatrix as SM
from SeqAli import SeqAli


def needleman_wunsch(sequence1, sequence2, similarity_matrix,
                     gap_init = None, gap_penalty = None):

    rows = len(sequence1) + 1
    cols = len(sequence2) + 1
    sequence1 = re.sub('[xX-]', '*', sequence1)
    sequence2 = re.sub('[xX-]', '*', sequence2)
    fMatrix = np.zeros((rows, cols), float)
    similarityMatrixMap = SM.get_matrix(similarity_matrix)

    if gap_init is None:
        gap_init = similarityMatrixMap['*']['A']
    if gap_penalty is None:
        gap_penalty = similarityMatrixMap['*']['*']

    for i in range(0, rows):
        fMatrix[i][0] = i * gap_init
    for j in range(0, cols):
        fMatrix[0][j] = j * gap_init

    for i in range(1, rows):
        for j in range(1, cols):
            mtch   = fMatrix[i - 1][j - 1]
            mtch  += similarityMatrixMap[sequence1[i - 1]][sequence2[j - 1]]
            delete = fMatrix[i - 1][j] + gap_penalty
            insert = fMatrix[i][j - 1] + gap_penalty
            fMatrix[i][j] = max(mtch, delete, insert)

    return trackBack(fMatrix, sequence1, sequence2,
                     gap_penalty, similarityMatrixMap)


def trackBack(fMatrix, seq1, seq2, gap, similarity_map):
    '''
    Tracks back to create the aligned sequence pair
    '''
    alignedSeq1 = ''
    alignedSeq2 = ''
    i = len(seq1)
    j = len(seq2)
    while i > 0 and j > 0:
        score = fMatrix[i][j]
        diagScore = fMatrix[i - 1][j - 1]
        upScore = fMatrix[i][j - 1]
        leftScore = fMatrix[i - 1][j]
        if score == diagScore + similarity_map[seq1[i - 1]][seq2[j - 1]]:
            alignedSeq1 = seq1[i - 1] + alignedSeq1
            alignedSeq2 = seq2[j - 1] + alignedSeq2
            i -= 1
            j -= 1
        elif score == leftScore + gap:
            alignedSeq1 = seq1[i - 1] + alignedSeq1
            alignedSeq2 = '-' + alignedSeq2
            i -= 1
        elif score == upScore + gap:
            alignedSeq1 = '-' + alignedSeq1
            alignedSeq2 = seq2[j - 1] + alignedSeq2
            j -= 1
        else:
            raise 'Not Possible'
    while i > 0:
        alignedSeq1 = seq1[i - 1] + alignedSeq1
        alignedSeq2 = '-' + alignedSeq2
        i -= 1
    while j > 0:
        alignedSeq1 = '-' + alignedSeq1
        alignedSeq2 = seq2[j - 1] + alignedSeq2
        j -= 1

    stats = alignment_stats(alignedSeq1, alignedSeq2, similarity_map)
    return SeqAli(sequences      = [alignedSeq1, alignedSeq2],
                  sequence_inits = [1, 1],  identities = stats['idt'],
                  positives = stats['pos'], gaps = stats['gaps'])


def alignment_stats(seq1, seq2, similarity_map):
    stats = {'idt': 0, 'pos': 0, 'gaps': 0}

    for i in range(len(seq1)):
        if seq1[i] == "-" or seq2[i] == "-":       stats['gaps'] += 1
        elif seq1[i] == seq2[i]:                   stats['idt']  += 1
        elif similarity_map[seq1[i]][seq2[i]] > 0: stats['pos']  += 1
    return stats



# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>

.. affiliation::
    Structural BioInformatics Lab <sbi.upf.edu>
    Baldo Oliva <baldo.oliva@upf.edu>

.. class:: ChainFrame
.. class:: ProteinChainFrame
.. class:: NucleotideChainFrame
"""

# Standard Libraries
import warnings

# External Libraries
import pandas as pd

# This Library
from SBILib.data import aminoacids_main3, nucleic_main
from ..Frame3D import Frame3D
from ..residue import ResidueFrame

__all__ = ['ChainFrame', 'ProteinChainFrame', 'NucleotideChainFrame']


class ChainFrame( Frame3D ):
    """A DataFrame container of single chain data. A protein fragment
    would also fall into this category.
    """
    _subtyp = 'chain_frame'

    def __init__( self, *args, **kw ):
        super(ChainFrame, self).__init__(*args, **kw)

    # Attributes
    @property
    def globalID( self ):
        """
        Full identifier of the Chain.
        By default, it would be ``<pdb-name>_<chain-id>``

        :returns: :py:class:`str`
        """
        return "{0}_{1}".format(self.id, self.chain)

    @property
    def pdb( self ):
        warnings.warn(".pdb will be deprecated. use .id instead", FutureWarning, stacklevel=2)
        return self.id

    @property
    def chain( self ):
        """
        Chain identifier.

        :returns: :py:class:`str`
        """
        return str(self._current_model[self._current_asym].unique()[0])

    # Magic Methods
    def __len__(self):
        """Number of residues in the Chain (of the first -main- entity of the Chain).
        """
        entity = self._current_model.iloc[0]["label_entity_id"]
        entity = self._current_model[self._current_model["label_entity_id"] == entity]
        return len(entity.groupby([self._current_seq, "pdbx_PDB_ins_code"]))

    # Private Pandas Methods
    def _inheritance( self, other ):
        self.id = other.id
        return self

    # Pandas Magic Methods
    @property
    def _constructor( self ):
        def f(*args, **kwargs):
            df = pd.DataFrame(*args, **kwargs)
            if len(df[self._current_seq].unique()) == 1:
                return ResidueFrame(*args, **kwargs).__finalize__(self, method='inherit')
            else:
                return ChainFrame(*args, **kwargs).__finalize__(self, method='inherit')
        return f

    def __finalize__(self, other, method=None, **kwargs):
        if method == 'inherit':
            if (self[self._current_comp].isin(aminoacids_main3) == True).any():
                return ProteinChainFrame(self).__finalize__(other, 'inherit')
            if (self[self._current_comp].isin(nucleic_main) == True).any():
                return NucleotideChainFrame(self).__finalize__(other, 'inherit')
            return self._inheritance(other)
        if method == 'concat':
            if len(self._current_model[self._current_asym].unique()) > 1:
                from SBILib.structure import PDBFrame
                return PDBFrame(self)
        return self


class ProteinChainFrame( ChainFrame ):
    _subtyp = 'protein_chain_frame'

    def __init__( self, *args, **kw ):
        super(ProteinChainFrame, self).__init__(*args, **kw)

    # Attributes
    @property
    def aminoacids( self ):
        """Iterate over the :class:`.ProteinResidueFrame` of the :class:`.ChainFrame`.

        :return: Iterator[:class:`.ProteinResidueFrame`]
        """
        pass

    # Pandas Magic Methods
    @property
    def _constructor( self ):
        def f(*args, **kwargs):
            df = pd.DataFrame(*args, **kwargs)
            if len(df[self._current_seq].unique()) == 1:
                return ResidueFrame(*args, **kwargs).__finalize__(self, method='inherit')
            else:
                return ProteinChainFrame(*args, **kwargs).__finalize__(self, method='inherit')
        return f

    def __finalize__(self, other, method=None, **kwargs):
        if method == 'inherit':
            return self._inheritance(other)
        else:
            return super(ProteinChainFrame, self).__finalize__(other, method, **kwargs)
        return self


class NucleotideChainFrame( ChainFrame ):
    _subtyp = 'nucleotide_chain_frame'

    def __init__( self, *args, **kw ):
        super(NucleotideChainFrame, self).__init__(*args, **kw)

    # Pandas Magic Methods
    @property
    def _constructor( self ):
        def f(*args, **kwargs):
            df = pd.DataFrame(*args, **kwargs)
            if len(df[self._current_seq].unique()) == 1:
                return ResidueFrame(*args, **kwargs).__finalize__(self, method='inherit')
            else:
                return NucleotideChainFrame(*args, **kwargs).__finalize__(self, method='inherit')
        return f

    def __finalize__(self, other, method=None, **kwargs):
        if method == 'inherit':
            return self._inheritance(other)
        else:
            return super(NucleotideChainFrame, self).__finalize__(other, method, **kwargs)
        return self
